"""
The console module
------------------

The classes defined in this module provide a command line interface to the
application.  The collection implements the frame interface.

Created on Mar 11, 2017
@author: neil
"""
import sys
from getpass import getpass
from ncexplorer.app import CmdApplication
from ncexplorer.frame.base import BaseFrame, BaseProgressBar
from ncexplorer.plotter.plotter import ScatterPlotter
from ncexplorer.plotter.basemapplotter import BasemapPlotter


# A text based progress bar.
#
# Thanks to Greenstick (http://stackoverflow.com/users/2206251/greenstick) for
# the printProgress() method.
class ProgressBarText(BaseProgressBar):
    
    # The interface object (GUI or console app), which is the frame object,
    # initializes the progress bar.  It is also the frame object that
    # understands the size constraints.  So the size parameters are set
    # here.
    def __init__(self, barLength=50):
        '''Initialize a text based progress bar.  The default length is 50.'''
        self._barLength=barLength
    
    def start(self, counts):
        '''
        Initialize a text base progress bar with a header.
        '''
        self._counts = counts
        self._prefix = 'Progress'
        self._suffix = 'Complete'
        self._i = 0
        
        self._printProgress(0, counts,
                            prefix=self._prefix,
                            suffix=self._suffix,
                            barLength=self._barLength)

    def _printProgress(self, iteration, total,
                       prefix='',
                       suffix='',
                       decimals=1,
                       barLength=100,
                       fill='|'):
        '''
        Call in a loop to create terminal progress bar
        @params:
            iteration   - Required  : current iteration (Int)
            total       - Required  : total iterations (Int)
            prefix      - Optional  : prefix string (Str)
            suffix      - Optional  : suffix string (Str)
            decimals    - Optional  : positive number of decimals in percent complete (Int)
            barLength   - Optional  : character length of bar (Int)
        '''
        percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
        filledLength = int(barLength * iteration // total)
        bar = fill * filledLength + '-' * (barLength - filledLength)
        sys.stdout.write('\r%s |%s| %s%s %s' % (prefix, bar, percent, '%', suffix)),
        if iteration == total:
            sys.stdout.write('\n')
        sys.stdout.flush()

    # FIX ME: Ignore the message for now.
    def update(self, msg):
        self._i += 1
        self._printProgress(self._i, self._counts,
                            prefix=self._prefix,
                            suffix=self._suffix,
                            barLength=self._barLength)

    # No closing operations required of a text-based progress bar.
    def close(self):
        pass

class ConsoleFrame(BaseFrame):
    """Does not seem worth it to write these docstrings.  They're not showing
    up in the generated documentation.  The examples I'm seeing all are
    writing the documentation directly into *rst files.
    """
    
    # Methods required to be implemented.
    def _display_matches(self, matches):
        self._matches = matches
        for i, repo, files in matches:
            print "{0}: {1}".format(i, repo.id)
            for j, filename in files:
                print "  {0} - {1}".format(j, filename)
        return matches

    # Choose the CmdApplication for the ConsoleFrame.
    def _set_application(self, title):
        app = CmdApplication(self, title)
        return app

    # The progress bar.  A bar length of 40 works with the Jupyter notebooks.
    # Not using the second parameter in the console.  The progress bar is just
    # a simple text implementation.
    def _set_progressbar(self):
        pb = ProgressBarText(barLength=20)
        return pb

    def _display_variables(self, payload):
        """Print each dataset on a line.
        
        Print each dataset on a line with its variables underneath and
        indented.
        """
        i = 0
        for ds in payload:
            dsline = "{0}: Institute: {1}, Model: {2}, Experiment: {3}".format(
                ds['ID'], ds['Institute'], ds['Model'], ds['Experiment'])
            print dsline

            for var in ds['Variables']:
                varline = "  -> {0} ({1}): ".format(
                    var['name'], var['longname'])
                print varline
            i += 1

    def _plotter(self, **kwargs):
        if 'charttype' in kwargs:
            charttype = kwargs['charttype']
            if charttype.upper() == 'SCATTER':
                plotter = ScatterPlotter()
            elif charttype.upper() == 'MAP':
                plotter = BasemapPlotter(**kwargs)
            else:
                msg = "Unrecognized chart type: {0}.".format(charttype)
                raise RuntimeError(msg)
        else:
            plotter = BasemapPlotter(**kwargs)

        # Let the plotting instance have a reference to this.  The plotting
        # instance needs the figure on the canvas managed by this.
        plotter.set_canvas(self._canvas)
        plotter.set_defaults()
        
        # Launch the canvas window.
        self._canvas.show()
            
        return plotter

    def get_login_creds(self):
        username = raw_input('Username: ')
        password = getpass('Password: ')
        return {'username': username, 'password': password}

    # The console implementation of this presumes that the user specified
    # the intended search parameters at job.search(...).  No need to prompt
    # again.
    def get_search_params(self):
        """Returns empty search string."""
        return ""

    # This is a convenience method for the Console app only.
    def new_plotter(self, **kwargs):
        """Create a new canvas with a single scatter or basemap figure.
        
        This method is only available on the ConsoleFrame."""
        if 'charttype' in kwargs:
            charttype = kwargs['charttype']
            if charttype.upper() == 'SCATTER':
                plotter = ScatterPlotter()
            elif charttype.upper() == 'MAP':
                plotter = BasemapPlotter(**kwargs)
            else:
                msg = "Unrecognized chart type: {0}.".format(charttype)
                raise RuntimeError(msg)
        else:
            plotter = BasemapPlotter(**kwargs)
        
        canvas = self._new_canvas()
        plotter.set_canvas(canvas)
        plotter.set_defaults()
        canvas.show()
        return plotter

    # Nothing implemented here yet.  A Ncurses terminal interface might be
    # helpful, but that's not trivial to implement.  Short of something like
    # that, the interactive shell is the best.
    def mainloop(self):
        """No mainloop implemented.
        
        Later, build an ncurses interface.
        """
        raise NotImplemented("No console based interface implemented. " +
                             " Use the interactive shell and import " +
                             "the class.")



"""
The frame module is the foundation of the user interface.  The parent class is
the BaseFrame object, which defines the user interface.  User interfaces
(e.g., a web interface, a desktop GUI or a console curses interface) implement
this interface.
"""


'''
Created on Jan 24, 2017

@author: neil
'''
import Tkinter as tk

# The import of ttk must follow tk, to override the basic tk widgets.
from ttk import Treeview, Button, Entry, Progressbar, Frame, Label

# The Tkinter backend for matplotlib.
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
#from matplotlib.figure import Figure

#from urlparse import urlparse
#import getpass

from ncexplorer.app import GUIApplication
from ncexplorer.plotter.basemapplotter import TkPlotter
from ncexplorer.frame.base import BaseFrame, BaseProgressBar
#from nc.expframe import parse_params, parse_variable_names


# The coordinates defining the globe's display.
class GlobeCoords(object):
    def __init__(self, center=[0,0], time_=0, plev=0):
        self.center = center
        self.time_ = time_
        self.plev = plev

    def __eq__(self, other):
        '''Equal if each component is equal by value.'''
        if isinstance(other, self.__class__):
            cond1 = (self.center == other.center)
            cond2 = (self.time_ == other.time_)
            cond3 = (self.plev == other.plev) 
            return (cond1 and cond2 and cond3)
        else:
            return NotImplemented

    def __ne__(self, other):
        '''Not equal if any component is different by value.'''
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        else:
            return NotImplemented

    def copy(self, other):
        '''Copy the values of components of the other.'''
        if isinstance(other, self.__class__):
            self.center = other.center
            self.time_ = other.time_
            self.plev = other.plev
            return self
        else:
            return NotImplemented

            
# The GUI is laid out as shown in the diagram.
# +---------------------------------------------------------------+
# |                                 |                             |
# |                                 |                             |
# |                                 |           SEARCH            |
# |                                 |                             |
# |                                 |                             |
# |                                 |                             |
# |           PLOT                  |-----------------------------|
# |                                 |                             |
# |                                 |                             |
# |                                 |           DATA              |
# |                                 |                             |
# |                                 |                             |
# |                                 |                             |
# +---------------------------------------------------------------+
GUI_HEIGHT=600
PLOT_WIDTH=600
CONTROLS_WIDTH=600
SEARCH_HEIGHT=300
VARS_HEIGHT=300

 
class PlotterFrame(Frame):
    '''
    A Tkinter sublcassed frame the provides Tk scale widgets to the Basemap
    plotting canvas to rotate the globe and update the time and plev
    dimensions.
    '''
    def __init__(self, parent):
        Frame.__init__(self, parent)
        
        fleft = Frame(master=self)
        fright = Frame(master=self)
        fleft.pack(side=tk.LEFT)
        fright.pack(side=tk.RIGHT)
        
        frtop = Frame(master=self)
        frmiddle = Frame(master=self)
        frbottom = Frame(master=self)
        frtop.pack()
        frmiddle.pack()
        frbottom.pack()
        
        # Define an object for _plotter, but create the scale widgets first.
        # Protection against the scale objects attempting to update the
        # plotter before it's created.
        self._plotter = None
        
        # The current time, plev and [lat, lon] point specified by the
        # scale widgets.
        self._coordinates = GlobeCoords(center=[30,-120])
        self._globecoords = GlobeCoords(center=[30,-120])

        # There are four scale objects: one each for latitude, longitude,
        # time and plev.
        self._scale_lat = tk.Scale(fleft,
                                label="Latitude",
                                from_=90,
                                to=-90,
                                resolution=30,
                                orient='vertical',
                                length=400,
                                command=self._save)
        self._scale_lon = tk.Scale(frbottom,
                                label="Longitude",
                                from_= -180,
                                to=180,
                                resolution=30,
                                orient='horizontal',
                                length=400,
                                command=self._save)
        self._scale_time = tk.Scale(frtop,
                                    label="Time",
                                    resolution=1,
                                    orient='horizontal',
                                    length=400,
                                    command=self._save)
        self._scale_plev = tk.Scale(frtop,
                                    label="Pressure Level",
                                    resolution=1,
                                    orient='horizontal',
                                    length=400,
                                    command=self._save)
        self._scale_lat.pack()
        self._scale_lon.pack()
        self._scale_time.pack()
        self._scale_plev.pack()
        
        # Set the values of the scales to self._coordindates, otherwise the
        # globe will jump on the first update.
        self._scale_lat.set(self._coordinates.center[0])
        self._scale_lon.set(self._coordinates.center[1])
    
        # The plotter
        self._plotter = TkPlotter(dpi=100,
                                  width=1.0*PLOT_WIDTH/100.0,
                                  height=1.0*GUI_HEIGHT/100.0)
        
        # This draws the empty globe.  It looks better than an empty white
        # square as the application is initializing.
#        self._plotter.clear()
        self._canvas = FigureCanvasTkAgg(self._plotter.figure, frmiddle)
#        self._canvas.show()
        self._canvas.get_tk_widget().pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

#        self._plotter = Plotter(self._axes, center=self._coordinates.center)

        self._poll()

    # Get all four values from the scale widgets and redraw the globe. 
    def _save(self, arg):
        self._coordinates.center = [self._scale_lat.get(), self._scale_lon.get()]
        self._coordinates.time_ = self._scale_time.get()
        self._coordinates.plev = self._scale_plev.get()

    # FIX ME:  The object organization for the plotter has changed.  This
    # is just a patch.  This module needs to be written following the new
    # paradigm.
    def draw(self, time_=0, plev=0, **kwargs):
        self._plotter.draw(time_=0, plev=0, **kwargs)
        
    def _draw(self):
        # The plotter object may not exist.
        if None != self._plotter:
            
            # If we have the variable, then include it.
            try:
                self._plotter.draw(self._axes,
                                   center=self._coordinates.center,
                                   var=self._variable,
                                   time_=self._coordinates.time_,
                                   plev=self._coordinates.plev)

            # Otherwise show the empty globe.
            except AttributeError:
                self._plotter.draw(self._axes,
                                   center=self._coordinates.center)
        
        self._canvas.draw()

    def _poll(self):
        '''Redraws the Basemap globe if any coordinate values have changed.'''
        if self._coordinates != self._globecoords:
            self._draw()
            self._coordinates.copy(self._globecoords)
        self.after(500, self._poll)

    def plotdata(self, var, time_=0, plev=0):
        # The variable var is an xarray variable.
        self._variable = var
        
#        # The time and plev scales need to be adjusted to match the number of
#        # these values in the variable.
#        newto_time = len(var['time'])
#        newto_plev = len(var['plev'])
#        self._scale_time.configure(to=newto_time)
#        self._scale_plev.configure(to=newto_plev)
#        self._draw()
        self._plotter.plotdata(var[0])

class ProgressUpdater(Frame, BaseProgressBar):
    '''
    A Tkinter subclassed widget that presents a progress bar and a message to
    inform the user.
    '''
    def __init__(self, parent, prompt):
        Frame.__init__(self, parent)

        # Calls the parent's update() method.
        self._parent = parent

        # The label is static.  It's just a prompt.  The prompt, the progress
        # bar and the message string are laid out in one row.  E.g.:
        # Prompt: |||||||...........  Updating 1 of 7.
        lbl = Label(self, text=prompt, background='')
        lbl.pack(side=tk.LEFT)

        self._progressbar = Progressbar(self, orient='horizontal')
        self._progressbar.pack(side=tk.LEFT)

        self._message = tk.StringVar()
        msg = Label(self, textvariable=self._message, width=60)
        msg.pack(side=tk.LEFT)

        # FIX ME: Not sure what to do about a default step size.
        self._setsize = 1 

    def update(self, msg=""):
        '''Write the message and advance the progress bar.'''
        self._progressbar.step(self._stepsize)
        self._message.set(msg)
        self._parent.update()

    # This presumes the length is 100, which is the default.
    def start(self, total_steps):
        # The progress bar adds 2 to the total number of steps.  1 is to
        # update the bar at the beginning, before anything is done.  2 is to
        # update the bar at the end.
        self._stepsize = 100.0/(total_steps + 2)
        self.update("")

    # Clearn up.
    def close(self):
        self._setsize = 1
        self.update("")
        self._parent.update()


class SearchFrame(Frame):
    '''
    Create the search cluster of widgets.  The cluster comprises a
    ProgressUpdater widget, a search entry box, a search button to initiate
    the search, and the tree view of URLs that have been found.
    '''
    def __init__(self, parent, handler):
        Frame.__init__(self, parent)

        # The parent's handler for the search button.
        self._parent_handler = handler
        
        # The label and progress bar
        ftop = Frame()
        ftop.pack(anchor=tk.W)
        self.progress_updater = ProgressUpdater(ftop, "Search: ")
        self.progress_updater.pack()

        # The search string and button to launch search.
        fmiddle = Frame()
        fmiddle.pack(anchor=tk.W)
        Button(fmiddle, text="Search", command=self._handler).pack(side=tk.LEFT)
        self._entry = Entry(fmiddle, width=60)
        self._entry.pack(side=tk.LEFT)

        # The tree view lists the source of data found from the search.
        fbottom = Frame()
        fbottom.pack(fill=tk.BOTH, expand=True)
        self._tree = Treeview(fbottom)
        self._tree.pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)
        self._tid = self._tree.insert('', index='end', text='Sources')

    def _handler(self):
        search_str = self._entry.get()
        self._parent_handler(search_str)

    def insert(self, dsline):
        self._tree.insert(self._tid, 0, text=dsline)


class VariableFrame(Frame):
    '''right
    Create the variable cluster of widgets.  The cluster comprises a
    ProgressUpdater widget, a retrieve button to retrieve the selected files,
    a tree view of the variables, and a plot button to plot the data.
    '''
    def __init__(self, parent, retriever, plotter):
        Frame.__init__(self, parent)

        # The parent's handlers for retrieving data and plotting the
        # variables.
        self._retriever = retriever
        self._plotter = plotter
        
        # The label and progress bar
        ftop = Frame()
        ftop.pack(anchor=tk.W)
        self.progress_updater = ProgressUpdater(ftop, "Retrieve: ")
        self.progress_updater.pack()

        # The retrieve and plot buttons.
        fbottom = Frame()
        fbottom.pack()
        fbleft = Frame()
        fbleft.pack(side=tk.LEFT, anchor=tk.N)
        Button(fbleft, text="Get Data", command=self._retriever).pack()
        Button(fbleft, text="Plot Data", command=self._plotter).pack()

        # The tree view lists the variables that match the variable supplied
        # in the initial search.
        fbright = Frame()
        fbright.pack(fill=tk.BOTH, expand=True)
        self._tree = Treeview(fbright)
        self._tree.pack(fill=tk.BOTH, expand=True)
        self._tid = self._tree.insert('', index='end', text='Variables')

    def insert(self, dsline, item_id=None, tags=None):
        if None == item_id:
            ret_id = self._tree.insert(self._tid, 0, text=dsline)
        else:
            ret_id = self._tree.insert(item_id, 0, text=dsline, tags=tags)
        return ret_id

    def selection(self):
        item = self._tree.selection()
        ret = self._tree.item(item)
        return ret


class TkinterFrame(BaseFrame):
    '''width=CONTROLS_WIDTH, height=GUI_HEIGHT
    GUI Window made with the Tkinter library.
    '''
    def __init__(self, title):
#        self._app = app
        self._title = title
        self._root = tk.Tk()
        self._root.wm_title(self._title)
        
        # Layout the screen.
        fr_plot = Frame()
        fr_right = Frame()
        fr_plot.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        fr_right.pack(side=tk.RIGHT)
        
        fr_rtop = Frame()
        fr_rbottom = Frame()
        fr_rtop.pack(side=tk.TOP)
        fr_rbottom.pack(side=tk.BOTTOM)
        
        self._plotter = PlotterFrame(fr_plot)
        self._plotter.pack()

        # The search area:
#        fr_search = Frame(master=fr_right, width=CONTROLS_WIDTH, height=SEARCH_HEIGHT)
#        fr_search.pack()
        self._fr_search = SearchFrame(fr_rtop, self._search_handler)
        self._fr_search.pack()
        
        # The variables area:
#        fr_vars = Frame(master=fr_right, width=CONTROLS_WIDTH, height=VARS_HEIGHT)
        self._fr_vars = VariableFrame(fr_rbottom, self._get_data, self._plot_handler)
        self._fr_vars.pack()

        # The parent's __init__ still required.
        BaseFrame.__init__(self, title)

    # Methods required to be implemented by the subclass.
    def _set_application(self, title):
        app = GUIApplication(self, title)
        return app

    def _set_progressbar(self):
        ret = self._fr_search.progress_updater
        return ret

    def _set_plotter(self):
        return self._plotter

    def _display_matches(self, matches):
        for i, repo, files in matches:
            dsline = "Repository: {0}".format(repo.id)
            dsid = self._fr_search.insert(dsline)

    def _display_variables(self, payload):
        """List all the datasets in a tree view.
        
        List all the datasets and the variables and coordinates in a tree view.
        """
        i = 0
        for ds in payload:
            dsline = "Institute: {0}, Model: {1}, Experiment: {2}".format(
                ds['Institute'], ds['Model'], ds['Experiment'])
            dsid = self._fr_vars.insert(dsline)
            
            for var in ds['Variables']:
                varline = "  -> {0} ({1}): ".format(
                    var['name'], var['longname'])
                self._fr_vars.insert(varline, item_id=dsid, tags=str(i))

#    # The progress bar and message line.  This is used by the application
#    # while it's searching and retrieving data.
#    def progressbar(self, op):
#        if 'search' == op:
#            ret = self._fr_search.progress_updater
#            return ret
#        if 'vars' == op:
#            ret = self._fr_vars.progress_updater
#            return  ret
#        raise RuntimeError("Expected 'search' or 'vars'.")

#    def progressbar_close(self, hndl):
#        # FIX ME: How to clean this up.
#        pass
#
    def _get_data(self):
        self.bind([(2,1)])

    def _search_handler(self, search_str):
        self.search(searchstr=search_str)
#        params = parse_params(search_str)
#        self._app.search(**params)

    def _plot_handler(self):
        self.plotdata('0:SNOWHICE')
#        item = self._fr_vars.selection()
#        if 'tags' in item:
#            tags = item['tags']
#            name = tags[0]
#            
#            # FIX ME: Can't get get the dataset that is selected, just the
#            # name of the variable.  So for, plotting just the first dataset.
#            ds = self._app.datasets[0]
#            var = ds[name]
#            self._app.plot(var)

#    def display_urls(self, urls):
#        '''
#        List all the URLS found from a search in the given repository.
#        '''
#        # FIX ME: Parsing should be done elsewhere.
#        for url in urls:
#            o = urlparse(url)
#            filename = o.path.split('/')[-1]
#            dsline = "Server: {0}, file: {1}".format(o.netloc, filename)
#            dsid = self._fr_search.insert(dsline)
            
#    def display_variables(self, datasets):
#        '''
#        List all the datasets, the variables and coordinates in a tree view.
#        '''
#        for ds in datasets:
#            # FIX ME: Preferably the repository should prepare the list of
#            # variables for the tree view.  That's where knowledge about the
#            # datasets resides.  The tree view just displays the list, and
#            # lets the user select a subset.
#            # 
#            # For now, build a tuple of the form:
#            #    (dataset, variable)
#            # and use that as the tag.
#            dsline = "Institute: {0}, Model: {1}, Experiment: {2}".format(
#                ds.institute_id, ds.model_id, ds.experiment_id)
#            dsid = self._fr_vars.insert(dsline)
#            
#            for vkey in ds.data_vars:
#                var = ds[vkey]
#                names = parse_variable_names(var)
#                varline = "{0} ({1}): ".format(names['name'], names['longname'])
#                tag = (ds, names['name'])
#                self._fr_vars.insert(varline, item_id=dsid, tags=names['name'])

#    def plotdata(self, var, center, time_=0, plev=0):
#        # Ignore the center.
#        self._plotter.plotdata(var, time_=time_, plev=plev)

#    # Implements the get_login_creds() interface.
#    # FIX ME: This should be a model dialog window.  However, Tkinter does not
#    # have a stock implementation for that.  There are examples of how to
#    # create a model dialog window on the internet, but they all involve
#    # 100 to 200 lines of code code, and none of them work as is. 
#    def get_login_creds(self):
#        '''Retrieve a username and password from the command line.'''
#        print "Login required"
#        username = raw_input('Username: ')
#        password = getpass.getpass('Password: ')
#        ret = {'username': username, 'password': password}
#        return ret

    def mainloop(self):
        self._root.mainloop()



'''
Created on Mar 17, 2017

@author: neil
'''
import unittest

from ncexplorer.frame.base import BaseFrame
from ncexplorer.app import parse_params

class TestProgressbar(object):

    def update(self, msg=""):
        pass
    def start(self, total_steps):
        pass
    def close(self):
        pass

class TestFrame(BaseFrame):
    
    def _display_matches(self, matches):
        self.matches = matches
    def _set_progressbar(self):
        pb = TestProgressbar()
        return pb


# Tests of all the acceptable search parameter syntaxes.
class TestSearchParamSyntax(unittest.TestCase):

    def test_no_space(self):
        self.assertEqual(
            {'key1': 'val1', 'key2': 'val2'},
            parse_params("key1=val1,key2=val2")
        )

    def test_extra_space(self):
        self.assertEqual(
            {'key1': 'val1', 'key2': 'val2'},
            parse_params("   key1    =    val1,    key2   =   val2   ")
        )

    def test_tab(self):
        self.assertEqual(
            {'key1': 'val1', 'key2': 'val2'},
            parse_params("key1    = val1,key2    = val2")
        )

    def test_single_quote(self):
        self.assertEqual(
            {'key1': 'val1', 'key2': 'val2'},
            parse_params("key1='val1', key2='val2'")
        )

    def test_space_in_val(self):
        self.assertEqual(
            {'key1': 'val 1', 'key2': 'val 2'},
            parse_params("key1 = 'val 1', key2 = 'val 2'")
        )
        
class TestSearchBase(unittest.TestCase):

    def test_baseframe_search_results(self):
        searchstr = ("experiment='lgm', time_frequency='monClim', "
                    + "institute='NCAR', variable='ta'")
        appframe = TestFrame('Test')
        matches = appframe.search(searchstr=searchstr)
        self.assertEqual(2, len(matches))


if __name__ == '__main__':
    unittest.main()


"""
Created on Nov 28, 2016

@author: neil
"""
import sys
import logging
import urllib2
from urlparse import urlparse
from ncexplorer.config import repositories
from repository import NCXESGF, NCXURS, LocalDirectoryRepository
from ncexplorer.util import simple_regrid


def parse_params(param_str):
    """
    Convert a string of the form name='value', ... into a dictionary.  Leading
    and trailing spaces are stripped, as are line-feed, carriage-return, tab,
    single-quote and double-quote characters.
    """
    params = {}
    for param in param_str.split(','):
        dirty = param.split('=')
        key = dirty[0].strip(' \t\n\r\'\"')
        value = dirty[1].strip(' \t\n\r\'\"')
        params.update({key: value})

    return params


# A dictionary like class for storing list data.
class NCDict(dict):
    
    def __init__(self, *args):
        dict.__init__(self, args)
    
    def __repr__(self):
        retstr = ""
        items = self.items()
        for item in items:
            key = item[0]
            value = item[1]
            retstr = retstr + "{0}: {1}\n".format(key, value)
        return retstr
            
    
class FileStore(object):
    """An object in which to save a list of files.
    
    This object is a temporary stub.  It should be replaced by a more robust
    object, perhaps from the collections library.
    
    This is a list of dictionaries of the form [{index, filename}] so that a
    collection of files can be represented like this:
        0: a_file.nc
        1: another_file.nc
    """
    def __init__(self):
        self._store = []
        self._next_i = 0

    def __getitem__(self, index):
        if type(index) is str:
            # FIXME: A collections object can support a more efficient
            # search method.
            for item in self._store:
                if item['filename'] == index:
                    return item['index'], item['filename']
            
            # At this point, the item wasn't found.
            msg = "{} not found.".format(index)
            raise IndexError(msg)

        elif type(index) is int:
            item = self._store[index]
            return item['index'], item['filename']
        else:
            msg = "The file request index must be an integer or a filename"
            raise TypeError(msg)
 
    def __len__(self):
        return len(self._store)

    def append(self, filename):
        entry = {'index': self._next_i, 'filename': filename}
        self._store.append(entry)
        self._next_i += 1

    def pretty_list(self):
        pretty_list = []
        for item in self._store:
            msgline = "{0}: {1}".format(item['index'], item['filename'])
            pretty_list.append(msgline)
        return pretty_list

    def __unicode__(self):
        pretty_list = self.pretty_list()
        return u'\n'.join(pretty_list)
    
    def __str__(self):
        return unicode(self).encode('utf-8')

    def __repr__(self):
        pretty_list = self.pretty_list()
        return u'\n'.join(pretty_list)

class MatchStore(object):
    """An object in which to store the matches of a search."""
    def __init__(self):
#        self._store = []
        self._store = {}
#        self._next_i = 0

    def __getitem__(self, index):
        # FIXME:  Use ordered dictionary.
        ks = list(self._store)
        key = ks[index]
#        item = self._store[index]
        item = self._store[key]
        return item['index'], item['repository'], item['files']

    def __len__(self):
        return len(self._store)

    def clear(self):
#        self._store = []
        self._store = {}
#        self._next_i = 0
#        self._next_j = 0

    def new_repo(self, repo):
        """Return the repo or create a new one if it doesn't exist."""
        if repo.id in self._store.keys():
            return repo.id

        files = FileStore()
#        entry = {'index': self._next_i, 'repository': repo, 'files': files}
        entry = {'index': repo.id, 'repository': repo, 'files': files}
#        self._store.append(entry)
        self._store[repo.id] = entry
#        self._next_i += 1
        return entry['index']

    def add_file(self, index, filename):
        repo = self._store[index]
        repofiles = repo['files']
        repofiles.append(filename)

    def select(self, selections):
        """Marks entries in the store for download.
        
        The input parameter selections takes the form [(i,j),...].  The second
        entry may be an integer, interpreted as the index.  Or the second entry
        may be a string, interpreted as the filename. 
        """
        # TODO: This can be made more efficient.
        self._selections = MatchStore()
        for repo_index, file_index in selections:
            
            # The first entry must be a string.  Raise a helpful exception if
            # this is not the case.
            if type(repo_index) is not str:
                msg = ("First enrty of a match request must be a string " +
                       "giving the repository's short name.")
                raise TypeError(msg)
                
            item = self._store[repo_index]
            repo = item['repository']
            files = item['files']
            i, filename = files[file_index]
            handle = self._selections.new_repo(repo)
            self._selections.add_file(handle, filename)

        # To avoid calling the getter selections() immediately after making
        # the selections, it can be returned here.
        return self._selections

    def selections(self):
        return self._selections

    def __str__(self):
        pretty_list = []
        for item in self._store:
            msgline = "{0}: {1}".format(item['item'], item['repository'])
            pretty_list.append(msgline)
            fstore = item['files']
            msgsubline = fstore.repr()
            pretty_list.append(msgsubline)
        return u'\n'.join(pretty_list)

# The master application
class Application(object):
    """
    The base application class.
    """
    def __init__(self, frameobj, title):
        self.title = title

#        # The logger.  Each subclass is responsible for the handler.
        self._logger = logging.getLogger(self.title)
        self._logger.setLevel(logging.INFO)
        self._add_log_handler(self._logger)
        self._logger.info(self.title + " starting.")

        # Let the subclass choose the platform for interacting with the user
        # (i.e., GUI or console).  The frame also instantiates a logger for
        # log like updates.
#        self._init_frame(self.title)
        self._frame = frameobj

        # Repositories can be servers that support OpenDAP, or local
        # directories of NetCDF files.
        # TODO: Check runtime if ESGF is supported, and expect that there
        # may be more than one server 
        self.repositories = {}
        for repospec in repositories:
            if repospec['type'] == 'esgf':
                repo = NCXESGF(repospec)
                self._add_repository(repo)
            elif repospec['type'] == 'urs':
                repo = NCXURS(repospec)
                self._add_repository(repo)
            elif repospec['type'] == 'local':
                repo = LocalDirectoryRepository(repospec)
                self._add_repository(repo)
            else:
                msg = ("Encountered unrecognized repository type {0} when " +
                       "adding repositories.".format(repospec['type']))
                self._logger.error(msg)

        # FIX ME: Still not sure about this.  These will probably be
        # references to OpenDAP resources, but not the data stored in memory.
        # A collection of datasets can be very large (~100s of Gb).
        #
        # However, because regridding (FIX ME:  Need reference to more
        # detail on this subject) can take many hours to accomplish, the
        # application will need to be able to save this work.
        self.datasets = {}

        # The variables.
        self.variables = {}
        
        # The repositories and datafile that meet the criteria of a search
        # are stored here.
        self._search_matches = MatchStore()

    # This method must be overriden by the subclass.
    def _add_log_handler(self, log):
        pass

#    # In some instances, e.g., the flask server, the frame needs to be known
#    # to another application.
#    def frame(self):
#        return self._frame

    # Only repository supported is ESGF.
    def _add_repository(self, repo):
        """Add the repository to the dictionary of repositories."""
        # The repository needs to be able to call this objects methods.
        # Setting the app property of the repository will "steal" the
        # repository, meaning if the repository's app property is something
        # else, this call will overwrite it.
        repo.set_app(self)
        self.repositories[repo.id] = repo

    def list_usernames(self):
        """Returns a list of the usernames associated with the repositories."""
        usernames = NCDict()
        for repo_id, repo in self.repositories.iteritems():
            usernames[repo_id] = repo.username()
#            usernames.append({repo_id: repo.username()})
        return usernames

    # FIXME:  This is hardcoded to /home/neil/workspace/data/neil
    def save(self, request):
        """Saves a file to repository.
        
        Accepts a list of tuples of the form (repo, file).  Then calls the
        corresponding repo's save() method with file."""
        for repo_id, ds in request:
            repo = self.repositories[repo_id]
            progressbar = self._frame.progressbar('push')
            repo.push(progressbar, ds)

    # FIX ME: This method would be more robust, and easier to test if it did
    # not presume the existence of a progress bar.  Better if it could work
    # (and return) without a progress bar.
    def search(self, **kwargs):
        """Perform a search for matching NetCDF files.

        Accepts a search string and then searches all supported repositories
        for files that match the parameters of the search string.  The search
        string may be of the form::

            searchstr="experiment='lgm', variable='ta'"

        where the single (or double quotes) surrounding the parameter's value
        are optional.  The search parameters may also be a comma-separated
        list of keyword arguments::

            experiment='lgm', variable='ta'

        The only parameter that is required is ``variable``.
        """
        # In some cases, unit testing most notably, there might not be a
        # progress bar.
        try:
            progressbar = self._frame.progressbar('search')
        except NameError:
            progressbar = None
            self._logger.info("Performing a search without a progress bar.")

        if 'searchstr' in kwargs:
            if len(kwargs['searchstr']) > 0:
                params = parse_params(kwargs['searchstr'])
            else:
                params = {}
        else:
            params = kwargs

        # TODO: If the kwargs is not of the first type, the second type is
        # naively presumed.  This is weak.
        if len(self.repositories) == 0:
            msg = ("Repositories must be added before a search can be performed.")
            raise RuntimeError(msg)

        # If no search parameters are provided, then ask the frame to prompt
        # the user for them.
        if len(params) == 0:
            searchstr = self._frame.get_search_params()
            if len(searchstr) > 0:
                params = parse_params(searchstr)
            else:
                params = {}

#        # A variable must be passed.
#        if 'variable' not in params:
#            msg = "'variable' not included in search parameters."
#            raise RuntimeError(msg)

        # FIX ME: Make the repositories list an object that can return
        # different iterators depending on whether local_only is true or
        # false.
        for repo in self.repositories.values():
            repo.set_search_params(**params)
            
            # Sometimes a server is down.  Therefore catch the URLerror.
            try:
                repo.search(self._logger, progressbar=progressbar)
            except urllib2.URLError, e:
                print e

        if progressbar is not None:
            progressbar.close()

    def search_results(self):
        """Return the results of a search as a list URLs"""
        self._search_matches.clear()
        for repo in self.repositories.values():
            handle = self._search_matches.new_repo(repo)
            for url in repo.urls():
                try:
                    urlobj = urlparse(url)
                except StandardError:
                    if url is None:
                        msg = ("Repository {0} returned None for urls()."
                               ).format(handle)
                    else:
                        msg = ("Could not parse {0} from repository {1}."
                               ).format(url, handle)
                    print msg
                filename = urlobj.path.split('/')[-1]
                self._search_matches.add_file(handle, filename)
        return self._search_matches

    def bind_data(self, request):
        """Creates xarray dataset objects from the request.

        Cycles through the repositories, downloading the selected files, and
        call the frame's display_variables method to display to the user the
        variables contained in the files.
        
        Parameters
        ----------
        request : list of tuples
            The first entry in the tuple is the repository index, which is
            displayed in the output of the application's search() method.  This
            is a string, corresponding to the short name of the repository.
            The second entry in the tuple can be the integer index of the
            specific file or the filename.  Both of these are listed in the
            application's search() method.
        """
        # This can take a while, especially since it depends on external
        # servers and the internet.
        progressbar = self._frame.progressbar('vars')

        ds_index = 0
#        for repo in self.repositories.itervalues():
        selections = self._search_matches.select(request)
        for i, repo, files in selections:

            # Retrieving the data depends on the success of making an HTTP
            # connection, logging on, and completing the conversation with the
            # server.  When it fails, the reason it fails should be displayed
            # to the user as a property of the server (repository?).
            skipds = False
            try:
                datasets = repo.retrieve_data(self._logger, progressbar, files)
            except IOError as err:
                if self._error_is_openid(err):
                    self._logger.error(
                        "OpenID error:  Access failure on node {0}.".format(repo.id)
                    )
                else:
                    self._logger.error("IO error: {0}".format(err))
                skipds = True

            # FIX ME: The error needs to be somehow pushed to the user
            # interface.
            if not skipds:
                
                # The variable datasets can be a list or a generator.  Assume
                # it's a list if it's not a function.
                if callable(datasets):
                    self.datasets[ds_index] = datasets
                else:
                    for i in datasets:
                        self.datasets[ds_index] = i
                        ds_index += 1

        if len(self.datasets) > 0:
            self._frame.display_variables(self.datasets)
        progressbar.close()

    def variable(self, varindex):
        """
        Returns the variable from the application's collection of datasets
        specified by the index.
        """
        # The varindex is a string of this form: i:vname.  The left of the ':'
        # is the index in datasets dictionary, and vname is the variable name
        # as it's represented in the dataset.
        var_selection = varindex.split(':')
        index = int(var_selection[0])
        var = var_selection[1]

        # The Xarray.dataset attribute data_vars gives a dictionary object of
        # the variables.  The attribute variables does not.
        dataset = self.datasets[index]
        retvar = dataset.data_vars[var]
        return retvar

    # FIX ME: Find a better solution for authentication than making the
    # application rely on handling the password.  Someone with access to this
    # code will have access to the password.
    def set_login_creds(self):
        creds = self._frame.get_login_creds()
        return creds

    # The mainloop is trivial, or not implemented in all cases but a desktop
    # application.
    def run(self):
        """The main loop.
        
        Calls the frame's mainloop method."""
        self._frame.mainloop()

    # Periodically, files can't be down loaded because in the client gets
    # redirected to a server that doesn't recognize the OpenID contect.  The
    # server serves a page explaining the a login is required instead of the
    # requested file.  The client process this as if it were a NetCDF file
    # and the parser fails.
    def _error_is_openid(self, err):
        """Checks if the file retrieve failed due to an OpenID failure."""
        if "Access failure" in err.message:
            return True
        else:
            return False

    # Utility Functions.
    # This first method should be a method of the xarray object.
    def regrid(self, var, likevar=None):
        # This can take a while, especially if there is a lot of time data.
        progressbar = self._frame.progressbar('vars')
        newvar = simple_regrid(var, progressbar, likevar)
        return newvar


# The only purpose for a subclass of the Application class is to implement
# different logging functionality.
#
# A command-line application, for driving the application from a command line.
# This class is the intended target for Jupyter Notebooks, development and
# writing automation scripts.
class CmdApplication(Application):
    """An application suitable to the command line.

    This application class is convenient for development and testing.  It is
    also the intended application class for Jupyter Notebooks and bash-based
    automation scripts.
    """
#    def _init_frame(self, title):
#        frame = ConsoleFrame(self, title)
#        self._frame = frame

    def _add_log_handler(self, log):
        hdlr = logging.StreamHandler(sys.stdout)
        log.addHandler(hdlr)


class GUIApplication(Application):
    """An application suitable to a desktop GUI.

    This application class is the intended application companion to a GUI that
    is compiled and runs on the local desktop.
    """
#    def _init_frame(self, title):
#        frame = TkinterFrame(self, title)
#        self._frame = frame

    def _add_log_handler(self, log):
        hdlr = logging.StreamHandler(sys.stdout)
        log.addHandler(hdlr)



'''
Created on Dec 28, 2016

@author: neil
'''
import sys
import math
import numpy as np
import xarray as xr
from scipy.spatial import KDTree, Delaunay
from scipy.interpolate import LinearNDInterpolator

from pydap.client import open_url
from pydap.cas.urs import setup_session


# Thanks to hernamesbarbara on github for this function.
#
# Googling did not find a built-in function, nor a function in the Numpy and
# SciPy libraries that produced a cartesian product of lists, where each
# element is a list instead of a tuple.  This is odd, because all the
# interpolation functions in SciPy require the inputs to be in this form.
# Seeming, this obligates the user to create this function on his own.
def cartesian(arrays, out=None):
    '''
    Generate a cartesian product of input arrays.
    Parameters
    ----------
    arrays : list of array-like
        1-D arrays to form the cartesian product of.
    out : ndarray
        Array to place the cartesian product in.
    Returns
    -------
    out : ndarray
        2-D array of shape (M, len(arrays)) containing cartesian products
        formed of input arrays.
    Examples
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])
    '''

    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype

    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)

    m = n / arrays[0].size
    out[:,0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        cartesian(arrays[1:], out=out[0:m,1:])
        for j in xrange(1, arrays[0].size):
            out[j*m:(j+1)*m,1:] = out[0:m,1:]
    return out


# For a list of variables, each represented as an xarray DataArray, compute
# the longitude, latitude lattice (grid) that is the dimension-wise union of
# all the coordinate dimensions. 
def union_coords(variables):
    '''
    For each dimension (i.e. lat, lon) form the union of points over all
    variables.  This function presumes that the names of the dimensions are
    lat and lon.
    
    :return a dictionary of sorted lists, one for each dimension.
    '''
    lat = set([])
    lon = set([])
    for v in variables:
        lat = lat | set(v['lat'].values)
        lon = lon | set(v['lon'].values)

    ret = {'lon': sorted(list(lon)), 'lat': sorted(list(lat))}
    return ret


# The following function is a second attempt.  It accomplishes the regridding 
# by iterating on all four dimensions.  It assumes that the plev and time
# coordinates will be identical for all datasets -- obviously not a safe
# assumption.
#
# The method should be replace by the regridding API provided in ESMFPy at
# www.earchsystemmodeling.org (part of ESGF).  Presumably it's faster and more
# likely to be correct.  Also, the ESMFPy API supports parallelizing the
# regridding operation.
#
def regrid(variables, lattice='union', tl=None, pl=None):

    # At present, only one method of prescribing the destination lattice is
    # supported.  That is, the destination lattice if formed by the union over
    # all lattices in the input list of variables.
    if 'union' == lattice:
        c = union_coords(variables)
        coords_lat = c['lat']
        coords_lon = c['lon']
    
    # Simple, very low-resolution lattice for development and testing.
    # Gridlines are every 30 degrees apart.
    elif 'latlon_30' == lattice:
        coords_lat = np.linspace(-90, 90, 7)
        coords_lon = np.linspace(0, 330, 12)
        
    # A low-resolution grid for testing.  This is not so low that there are
    # only a few simplices, yet it's not too big so that it's reasonably
    # fast.
    elif 'latlon_5' == lattice:
        coords_lat = np.linspace(-90, 90, 37)
        coords_lon = np.linspace(0, 355, 72)

    elif 'latlon_1' == lattice:
        coords_lat = np.linspace(-90, 90, 181)
        coords_lon = np.linspace(0, 359, 360)

    else:
        errMsg = "Lattice %s is not supported." % (lattice)
        raise NotImplementedError(errMsg)

    # The destination lattice won't change.  Now is a good time to build the
    # indexing structure on it.
    coords = cartesian(( [coords_lat, coords_lon] ))
#    coordtree = KDTree(coords)

    # This point starts the process of regridding.
    newvars = []
    for var in variables:
        
        # Establish the levels for time and plev at which the variable is
        # interpolated on the new lattice.  The time to complete this procedure
        # scales linearly with the number of these variables.  Therefore, allow
        # the user to specify exactly what's needed.
        if tl == None:
            time_levels = np.arange(len(var['time'].values))
        else:
            time_levels = tl
        if pl == None:
            plev_levels = np.arange(len(var['plev'].values))
        else:
            plev_levels = pl

        # These variables are
        # all presumed to be defined on a four-dimensional grid: lon, lat,
        # plev, time.  It is further assumed that time and plev don't need
        # to be regrid, because all the variables are defined for the same
        # values of those dimensions.
        newdata = np.empty( (len(time_levels), len(plev_levels), len(coords_lat), len(coords_lon)) )
        newdata[:] = np.NaN
        
        # This aspect of Python cost me an afternoon.  if you create an
        # iterator like this:
        #  myiterator = enumerate(var['plev'].values[plev_levels])
        # ...then that iterator can only be used once.  Once the code
        # iterfaces through it's values, it's done.  So nested loops using
        # these DO NOT WORK.
        for t, time_ in enumerate(var['time'].values[time_levels]):
            for p, plev in enumerate(var['plev'].values[plev_levels]):
                
                # Impose a Delaunay triangulation and an indexing structure
                # on the variable.
                varcoords = cartesian(( [var['lat'], var['lon']] ))
                vartri = Delaunay(varcoords)

                # Flatten the variable on lat and lon.
                # CHECK THAT var[time,plev].values.flatten()
                # has the same shape (N,2) as varcoords
                # and that they are in the same order.
                x = var[t,p].values.flatten()
                
                # Loop on the destination lattice dimensions.
                numcoords = len(coords)
                print "[Time: %s][Plev: %s] %s(%s points)" % (time_, plev, '-'*int(numcoords/1000), numcoords)
                print "[Time: %s][Plev: %s] " % (time_, plev),
                
                tmp = np.empty(len(coords))
                for i, lattice_point in enumerate(coords):
                    n = vartri.find_simplex(lattice_point)
                    pts = vartri.simplices[n]
                    spts = vartri.points[pts]
                    
                    f = LinearNDInterpolator(spts, x[pts])
                    tmp[i] = f(lattice_point)
                    
                    if (i + 1)%1000 == 0:
                        sys.stdout.write(".")
                print ""
                
                # Reshape the tmp array to organize it by lat, lon.
                tmp = np.reshape( tmp, (len(coords_lat), len(coords_lon)) )
                newdata[t,p,:,:] = tmp
                
        # Package the new variable, defined on the destination lattice, into an
        # xarray DataArray.
        foo = xr.DataArray(newdata,
                           coords=[var['time'].values[time_levels],
                                   var['plev'].values[plev_levels],
                                   coords_lat,
                                   coords_lon],
                           dims=['time', 'plev', 'lat', 'lon'],
                           name=var.name,
                           attrs=var.attrs)

        newvars.append(foo)


    return newvars

def simple_regrid(var, progressbar, likevar=None):
    """Regrids to 1x1 a variable of lat, lon only"""

    if likevar is None:
        coords_lat = np.linspace(-90, 90, 72)
        coords_lon = np.linspace(0, 360 - 2.5, 144)
    else:
        # The lat and lon dimensions are inherited from the target lat-lon
        # grid.  The time dimension to be unchanged, so it is inherited from
        # the original var.
        # FIXME:  There may not be a time coordinate.
        coords_lat = likevar['lat'].values
        coords_lon = likevar['lon'].values
        coords_time = var['time'].values
    
    # Display the ranges involved in the regrid, for confidence.
    varlatr = "[{0},{1}]".format(
        var.lat.values[0], var.lat.values[len(var.lat) - 1])
    varlonr = "[{0},{1}]".format(
        var.lon.values[0], var.lon.values[len(var.lon) - 1])
    tgtlatr = "[{0},{1}]".format(
        coords_lat[0], coords_lat[len(coords_lat) - 1])
    tgtlonr = "[{0},{1}]".format(
        coords_lon[0], coords_lon[len(coords_lon) - 1])
    print "The regrid will by from {0} X {1} --> {2} X {3}.".format(
        varlatr, varlonr, tgtlatr, tgtlonr)

    # FIXME:  This needs to be a more general routine.  Testing for the range
    # being used needs to be more robust.
    # If the range of longitudes is (-180,180), take the variable apart and
    # rebuild it with the range (0,360).
    if var.lon.values[0] < -10.0:
        blankdata = np.empty( (len(var.time), len(var.lon), len(var.lat)) )
        blankdata.fill(np.NaN)
        blanklon = []
        usevar = None
        i = 0
        for l in var.lon.values:
            blanklon.append(math.fmod(l + 360, 360))
            blankdata[:,i,:] = var.isel(lon=i)
            i =+ 1
        usevar = xr.DataArray(blankdata,
                              coords={'time': var.time,
                                      'lon': blanklon,
                                      'lat': var.lat},
                              dims=['time', 'lon', 'lat'])
    else:
        usevar = var
        

    # Impose a Delaunay triangulation and an indexing structure
    # on the variable.
    varcoords = cartesian(( [usevar['lat'], usevar['lon']] ))
    vartri = Delaunay(varcoords)

    # Create a DataArray object with the new shape filled with np.NaNs.  This
    # will be filled with values interpolation from the variable var.
    blankdata = np.empty( (len(coords_lat), len(coords_lon), len(coords_time)) )
    blankdata.fill(np.NaN)
    newvar = xr.DataArray(blankdata,
                          coords={'lat': coords_lat,
                                  'lon': coords_lon,
                                  'time': coords_time},
                          dims=['lat', 'lon', 'time'])

    # Carry over the appropriate attributes.  The name attribute is not in the
    # attrs dictionary.  But it's necessary; it's the variables name.
    newvar.attrs = var.attrs
    newvar.name = var.name
    
    # Set the missing_value attribute to 'NaN (numpy)'.
    newvar.attrs['missing_value'] = 'NaN (numpy)'

    # Show progress at the latitudes.  Otherwise it's too many.
    progressbar.start( len(coords_lat)*len(coords_lon) )

    # Iterate the new variable's grid, calculating the value from an
    # interpolation of the simplices in the original variable.
    # FIX ME: The interpolation function is calculated for as many points of
    # the new variable that fall in a single simplex.  It only needs to be
    # calculated once per simplex.
    for i in range(0, len(coords_lat)):
        for j in range(0, len(coords_lon)):
        
            # These steps produce the i-th simplex (triangle) in the variable.
            lattice_point = [coords_lat[i], coords_lon[j]]
            simplex_number = vartri.find_simplex(lattice_point)
            if simplex_number >= 0:
                simplex_points = vartri.simplices[simplex_number]
                simplex = vartri.points[simplex_points]
                
                # These steps produce the n-values of the variable for the n-simplex.
                # Since the simplices are triangles, these steps produce three values.
                #
                # Do this for each value of time.
                for t in range(0, len(coords_time)):
                    vals = []
                    for k, vertex in enumerate(simplex):
                        val_t = usevar.sel(lat=vertex[0], lon=vertex[1])
                        val = val_t.isel(time=t)
                        vals.append(val)
            
                    # These steps produce the function f that calculates the interpolated
                    # value anywhere in the simplex (triangle), and then calculates that
                    # value at the lattice point, and assigns it.
                    f = LinearNDInterpolator(simplex, vals)
                    x = f(lattice_point)
                    newvar[i,j, t] = x 
        
            # Show progress.
            msg = "Calculated values for ({:2.3f}, {:2.3f}).".format(coords_lat[i], coords_lon[j])
            progressbar.update(msg)

    return newvar

def transpose_var(data):
    """Determines if the data in an Xarray DataArray is transposed.
    
    Occasionally, the data in a dataset is organized with the longitude
    coordinate first and the latitude coordinate second.  This is causing
    problems with the matplotlib contourf and addcyclic methods.
    
    This function determines the lon dimension by comparing the shape of the
    data to the lengths of the latitude and longitude coordinates.  If the
    shape is such that the longitude is first, the data is transposed and the
    new DataArray is returned.
    """
    data_shape = data.shape
    if data_shape[0] == len(data.lon.values):
        # Longitude is first.  It needs to be transposed.
        var_name = data.name
        var_data = data.values
        new_var_data = np.transpose(var_data)
        new_array = xr.DataArray(new_var_data,
                                 coords=[data.lat.values, data.lon.values],
                                 dims=['lat', 'lon'])
        new_array.name = var_name
        return new_array
    else:
        return data

def get_urs_file(url):
    username = 'godfrey4000'
    password = 'J#bunan0'
#    the_url = 'https://disc2.gesdisc.eosdis.nasa.gov:443/opendap/TRMM_L3/TRMM_3B42.7/1998/001/3B42.19980101.03.7.HDF'
    session = setup_session(username, password, check_url=url)
#    dataset = open_url(the_url, session=session)
    ds = xr.open_dataset(url, decode_cf=False, engine='pydap', session=session)
    return ds



'''
Created on Dec 21, 2016

@author: neil
'''
import os
import ntpath
from webob.exc import HTTPError
from pyesgf.logon import LogonManager
from pydap.cas.esgf import setup_session
from pyesgf.search import SearchConnection
import xarray as xr
import numpy as np
from datetime import datetime, timedelta
from dateutil import relativedelta
from urlparse import urlparse

from ncexplorer.config import CFG_ESGF_NODE
from ncexplorer.config import CFG_ESGF_SEARCH_NODE
from ncexplorer.config import CFG_ESGF_OPENID_NODE
from ncexplorer.config import TRIVIAL_USERNAME, TRIVIAL_PASSWORD
from ncexplorer.util import get_urs_file
from fileinput import filename


# The authenticator classes make it possible to use different authentication
# methods with different repositories.  It also allows for trivial
# authentication methods for use in testing and development.
class Authenticator(object):
    """The Authenticator handles login and logout to repositories."""
    def __init__(self, app):
        """
        The authenticator requires access to the parent application so that
        it can ask the application to prompt and collect the username and
        password.
        """
        self._username = None
        self._password = None
        self._lm = None
        self._app = app
        self.session = None

        # Setup the login manager.  This will depend on the repository.
        self._set_lm()

    def username(self):
        """Return the username"""
        return self._get_username()

    # This method must be overriden by the subclass.
    def _set_lm(self):
        pass
    def _get_username(self):
        return self._username

    def login(self):
        return False

    def logout(self):
        pass

# Standard class for ESGF authentication.
class OpenIDAuthenticator(Authenticator):
    """Authenticates to an ESGF node using open ID."""
    
    def _oid(self):
        return CFG_ESGF_OPENID_NODE + '/' + self._username

    # The OpenID has a long server name component and is easy to forget.
    # Making it accessible as a property is often useful.
    def _get_username(self):
        return self._oid()

    # Use the login manager from the pyesgf library.
    # NOTE: This library uses 'logon/logoff' everywhere.  The convention
    # used throughout this application is 'login/logout'
    def _set_lm(self):
        self._lm = LogonManager()
#        self._username = None
#        self._password = None

        # Ensure the user is logged out initially.
        self._lm.logoff()

    def login(self, url):
        if self._username is None or self._password is None:
            creds = self._app.set_login_creds()
            self._username = creds['username']
            self._password = creds['password']

        oid = self._oid()

        # The function setup_session only raises the most general exception,
        # so there's no choice by to catch the base exception class.
        try:
            self.session = setup_session(oid, self._password, check_url=url)
        except Exception as e:
            print e
            return False
        return True
        

    def logout(self):
        self._lm.logoff()


class URSAuthenticator(Authenticator):
    
    def login(self):
        pass
        
# Trivial authentication with hard coded user name and password.  It uses the
# OpenID authentication, but eliminates typing in the password a zillion
# times.
class TrivialAuthenticator(OpenIDAuthenticator):
    """Just for testing.  Do not use."""
    def _set_lm(self):
        self._lm = LogonManager()
        self._username = TRIVIAL_USERNAME
        self._password = TRIVIAL_PASSWORD
        self._lm.logoff()


# This class does no authentication at all.  It's used for local disk access.
class NullAuthenticator(Authenticator):
    """Performs no login function."""
    # The parent class returns False.  The intention is that login fails unless
    # the sub class implements a login capability.  In this class, login will
    # always work.  Presumably, permissions are set on the directory.
    def login(self):
        return True

    def logout(self):
        pass


# The parent class for all repositories..
class NCXRepository(object):
    """
    Interface class for searching archives of NetCDF data (repositories),
    retrieving data from the repository, and authenticating to the
    repository if required.
    """
#    shortname = None
    
    def __init__(self, repospec):

        self._app = None
        self._authenticator = None
        self._search_params = None
        self._urls = None

        # Set the ID.  This is a unique constant the serves as the key in the
        # applications dictionary of repositories.
        self._repo_type = repospec['type']
        self._repo_parameters = repospec['parameters']

        # The structure of the repository parameters is different for different
        # repository classes.  So the sub class sets the id.
        self.id = self._set_id()
        
#        # Repositories should set a short name.  If this hasn't been set, raise
#        # and exception so it can be addressed.
#        if self.shortname is None:
#            msg = ("Repository {}: attribute short name not defined.  " +
#                   "Repository subclasses must assign the shortname " +
#                   "attribute").format(self.id)
#            raise NameError(msg)

    def username(self):
        """Return the username used by the authenticator."""
        return self._authenticator.username()

    # Subclasses implementing this interface must implement these methods:
    def _set_id(self, repospec):
        '''Set the repository ID to a constant unique to repositories'''
        pass
    def _set_authenticator(self):
        """
        Set the class that will implement authentication  (logging in and
        logging out).
        """
        pass
    def _search(self, log, progressbar):
        pass
    def _retrieve_data(self, log, progressbar, files):
        pass

    # The repository will have occasion to call the parent application's
    # methods.  Any repository method that does this can't be called until
    # this method as set the _app property.
    def set_app(self, app):
        """Set the _app property to the parent application."""
        self._app = app

        # Let the subclass choose the authentication method.
        self._authenticator = self._set_authenticator()

    def push(self, progressbar, ds):
        """Push the dataset from the client to the repository."""
        return self._push(progressbar, ds)

    def retrieve_data(self, log, progressbar, files):
        """Retrieve the data specified in the saved OpenDAP URLS."""
        return self._retrieve_data(log, progressbar, files)

    # The search parameters are set here, at the level of the base class, in an
    # attempt to standardize the search parameters across repositories.  This
    # might be too ambitious.
    def set_search_params(self, **kwargs):
        """Set the search parameters for a search."""
        self._search_params = kwargs

    def search(self, log, progressbar=None):
        """Conduct a search using the preset search parameter."""
        self._search(log, progressbar)

    def urls(self):
        """Return the saved URLs"""
        return self._list_urls()
    
    def _list_urls(self):
        return self._urls


class NCXURS(NCXRepository):
    
    # A generator that returns a pydap proxy for all the files in the
    # directory of the URS NASA Earthdata server.
    def _genfiles(self):
        # The Earthdata starts in 1997 and goes to roughly the present.  The
        # range here stops at the end of 2016 so as to not introduce any bias
        # in seasons.
        onemonth = relativedelta.relativedelta(months=1)
        cur_dt = datetime.strptime('1998-01-01 00:00:00', '%Y-%d-%m %H:%M:%S')
        end_dt = datetime.strptime('2017-01-01 00:00:00', '%Y-%d-%m %H:%M:%S')
        while cur_dt < end_dt:
            
            # There is a peculiarity in the file naming convention.  The first
            # file of every day is named with with current day, but in the
            # directory for the previous day.
            fyear = cur_dt.year
            fmonth = cur_dt.month
                    
            filename = "%04d/3B43.%04d%02d01.7.HDF" % (
                fyear, fyear, fmonth)
            filename_a = "%04d/3B43.%04d%02d01.7A.HDF" % (
                fyear, fyear, fmonth)
            remotefile = (self._repo_parameters['server'] +
                          self._repo_parameters['directory'] +
                          filename)
            remotefile_a = (self._repo_parameters['server'] +
                          self._repo_parameters['directory'] +
                          filename_a)

            # Retrieve the file using the pydap library, implemented in the
            # util module.
            try:
                ds = get_urs_file(remotefile)
            except HTTPError:
                print "{0}: 404.".format(filename)
                try:
                    ds = get_urs_file(remotefile_a)
                except HTTPError:
                    print "{0}: 404.".format(filename_a)
                    cur_dt = cur_dt + onemonth
                    continue
            yield ds
            
            # Advance to the next day.
            cur_dt = cur_dt + onemonth

    def _set_id(self):
        # FIXME:  Tthis is a total kludge.
        return 'URS'
    
    def _set_authenticator(self):
        auth = URSAuthenticator(self._app)
        return auth
    
    def _search(self, log, progressbar):
        # This response simply adversises the form of the files that are
        # available on this repository.
        self._urls = ['3B43.[YYYYMMDD].[hh].7.HDF']
                        
    def _retrieve_data(self, log, progressbar, files):
        
#        temp_ds = []
#        for i, remotefile in files:
#            url = self.base_url + remotefile
#            ds = get_urs_file(url)
#            temp_ds.append(ds)
#        
#        return temp_ds
        return self._genfiles
        
            
# Implementation of the NCXRepository class for the ESGF servers.
class NCXESGF(NCXRepository):
    """Implements the NXSearch interface for ESGF."""
    
    def _set_id(self):
        if self._repo_type != 'esgf':
            msg = ("NCXESGF repository class cannot be created with a " +
                   "specification type {0}.".format(self._repo_type))
            raise TypeError(msg)

        node = self._repo_parameters['node']        
        return 'ESGF'

    # TODO: This is set to the trivial authenticator, waiting for a
    # means to collect user name and password from the web GUI and the
    # desktop GUI.
    def _set_authenticator(self):
        auth = TrivialAuthenticator(self._app)
#        auth = OpenIDAuthenticator(self._app)
        return auth

    # Build a list of OpenDAP URLs that that match the the search string
    # contained in self._search_params.
    def _search(self, log, progressbar):
        """Execute the search using the pyesgf library.
        
        Execute the search using the pyesgf library, which uses the ESGF
        search API.  Stores a list of OpenDAP URLs as resources.
        """
        # FIXME: This should be part of the initialization.
        # The urls() method must return a list.
        self._urls = {}

        # Can't allow an empty search string; the repository is too big and
        # will return too many files.  An empty search string is legal because
        # local repositories are small, and an empty string asks for all the
        # files.
        if len(self._search_params) == 0:
            return

        esgf_node = self._repo_parameters['search_node']
        conn = SearchConnection(esgf_node, distrib=True)
        ctx = conn.new_context(**self._search_params)
        hit_count = ctx.hit_count

        # Each search clears the files from before.  The pyesgf library allows
        # for searches to be refined.  Consider utilizing that capability here.
        if hit_count > 0:
            progressbar.start(hit_count)
#            self._variable = self._search_params['variable']
            datasets = ctx.search()
            i = 1
            for dsresult in datasets:
                if 'variable' in self._search_params:
                    remotefiles = dsresult.file_context().search(
                        variable=self._search_params['variable'])
                else:
                    remotefiles = dsresult.file_context().search()
                msg = "Searching %s of %s.  %s files." % (i, hit_count, len(remotefiles))
                log.debug(msg)
    
                for remotefile in remotefiles:
                    urlobj = urlparse(remotefile.opendap_url)
                    filename = urlobj.path.split('/')[-1]
                    self._urls[filename] = remotefile.opendap_url
                i += 1
                progressbar.update(msg)

    # Not permitted to push files to the ESGF repositoris.
    def _push(self, progressbar, ds):
        msg = "ESGF repository is read only.  Pushing data not permitted."
        raise TypeError(msg)

    def _retrieve_data(self, log, progressbar, files):
        """Retrieve data using the pyesgf library.

        Execute the search using the pyesgf library, which uses the ESGF
        search API.  Then retrieve the data and make it available to the
        application as an xarray Dataset object.
        """
        # ESGF frequently doesn't work.  Until I get a document from them
        # that specifies a reliable API, I'm giving up.
        msg = "ESGF has become too unreliable, so it's temporarily unsupported."
        raise NotImplementedError(msg)
#        login_successful = self._authenticator.login()
#        if not login_successful:
#            self._app.logger.warn("Failed to login.")
#        session = self._authenticator.session

        temp_ds = []
        url_length = len(files)
        session = None

        # Add two to the progress bar.  One for just starting, and another
        # for when it's all finished.  Without these extra, the user can be
        # looking at a blank progress bar for the whole time, since _clean()
        # takes so long.
        progressbar.start(2*url_length)
        for i, remotefile in files:

            # The remotefile is just the filename, which is nicer for display.
            # Need the full url.
            url = self._url_from_file(remotefile)
            if session is None and self._authenticator.login(url):
                session = self._authenticator.session
            
            if session is not None:
                xdataset = xr.open_dataset(url,
                                           decode_cf=False,
                                           engine='pydap',
                                           session=session)
                msg = "Cleaning: {0}.".format(remotefile)
#            # Normalize it.
#            # FIX ME: Consider moving this to another place.  This
#            # operation is the biggest bottleneck of this searching and
#            # retrieving data.
#                self._clean(x)

                temp_ds.append(xdataset)
                msg = "Retained: {0}".format(filename)
                log.debug(msg)    
                progressbar.update(msg)
                
            else:
                msg = "Login failed."
                print msg
            log.debug(msg)
            progressbar.update(msg)

        # Don't stay logged on.
        self._authenticator.logout()

        # Return the list of xarray Dataset objects.  The Data_repospecset data
        # structure can't hold the datasets thus far collected because, in
        # general, their coordinates will be defined on different lattices.
        return temp_ds

    # FIXME: 
    def _url_from_file(self, fn):
        return self._urls[fn]

    def _list_urls(self):
        return self._urls.values()

    # This method contains a body of routines for normalizing the data
    # coming from ESGF
    #     (1) Datasets supply a real number (typically 1e+20) that is to be
    #         interpreted a missing values.  This causes problems with
    #         plotting utilities and interpolation algorithms
    #     (2) The time values differ from dataset to dataset, while the
    #         intervals do not.  With different values, arithmetic on datasets
    #         fails.
    def _clean(self, dataset):
        """Cleans the dataset from ESGF.

        The dataset passed as a parameter is mutable so it gets modified.
        """
        # Replace missing values with numpy's NaN.  The missing value is
        # usually 1e+20, but values can be like 1.0000002e+20, which is
        # different.  Ergo the inequality.
        for var in dataset.data_vars.itervalues():
            if 'missing_value' in var.attrs:
                missing_data_value = var.missing_value
                var.values[var.values >= missing_data_value] = np.NaN

        # FIX ME: This should check if the time type is monClim.  Also, check
        # to see if the first value should be 15.5.  In the meantime, so that
        # the value is at the midpoint of the interval, shift everything by
        # half.
        time0 = dataset.coords['time'][0]
        delta = 0.5*(dataset.coords['time'][1] - dataset.coords['time'][0])
        dataset.coords['time'] -= time0


class LocalDirectoryRepository(NCXRepository):
    """A repository consisting of NC files on the local disk

    This repository is intended primarily for testing and unit tests.
    """
    def __init__(self, repospec):
        self._path = None
        NCXRepository.__init__(self, repospec)

    # The sublass must implement the repository interface.
    def _set_id(self):
        if self._repo_type != 'local':
            msg = ("LocalDirectoryRepository class cannot be created with " +
                   "a specification type {0}.".format(self._repo_type))
            raise TypeError(msg)

        path = self._repo_parameters['path']
        self._path = path
        
        # Use only the directory name for the shortname because it's typed by
        # the user when requesting files.
        id = self._path_leaf(path)
#        self.shortname = sname

        return id

    def _path_leaf(self, path):
        head, tail = ntpath.split(path)
        return tail or ntpath.basename(head)

    def _set_authenticator(self):
        auth = NullAuthenticator(self._app)
        return auth

    def _search(self, log, progressbar):
        """Collects the output of ls -1 *nc in the repository directory."""
        self._urls = []
        for filename in os.listdir(self._path):
            url = 'file:////' + filename
            self._urls.append(url)
        self._urls.sort()

    def _push(self, progressbar, ds):
        # The filename comes from the data.
        filename = ds.filename
        filespec = self._path + '/' + filename
        ds.to_netcdf(filespec)
        
    def _retrieve_data(self, log, progressbar, files):

        # Add two to the progress bar.  One for just starting, and another
        # for when it's all finished.  Without these extra, the user can be
        # looking at a blank progress bar for the whole time, since _clean()
        # takes so long.
        temp_ds = []
        file_len = len(files)
        progressbar.start(2*file_len)
        for i, localfile in files:

            # FIX ME: Explore cheaper ways to determine if the file contains
            # the requested variable.  This implementation downloads the
            # entire file, and then discards it.
            xdataset = xr.open_dataset(self._path + '/' + localfile, decode_cf=False)
#            if self._search_params['variable'] in xdataset:

            urlobj = urlparse(localfile)
            filename = urlobj.path.split('/')[-1]
            msg = "Cleaning: [{0}] {1}.".format(urlobj.netloc, filename)
            log.debug(msg)
            progressbar.update(msg)

            # Normalize it.
            # FIX ME: Consider moving this to another place.  This
            # operation is the biggest bottleneck of this searching and
            # retrieving data.
#                self._clean(x)

            temp_ds.append(xdataset)
            msg = "Saved: [{0}] {1}".format(urlobj.netloc, filename)
            log.debug(msg)
            progressbar.update(msg)

        return temp_ds


"""
Taylor diagram (Taylor, 2001) test implementation.

http://www-pcmdi.llnl.gov/about/staff/Taylor/CV/Taylor_diagram_primer.htm

Thanks to ycopin on github for the first version of this code.
https://gist.github.com/ycopin/3342888
"""

__version__ = "Time-stamp: <2012-02-17 20:59:35 ycopin>"
__author__ = "Yannick Copin <yannick.copin@laposte.net>"

import numpy as NP
import matplotlib.pyplot as PLT

class TaylorDiagram(object):
    """Taylor diagram: plot model standard deviation and correlation
    to reference (data) sample in a single-quadrant polar plot, with
    r=stddev and theta=arccos(correlation).
    """

    def __init__(self, refstd, fig=None, rect=111, label='_'):
        """Set up Taylor diagram axes, i.e. single quadrant polar
        plot, using mpl_toolkits.axisartist.floating_axes. refstd is
        the reference standard deviation to be compared to.
        """

        from matplotlib.projections import PolarAxes
        import mpl_toolkits.axisartist.floating_axes as FA
        import mpl_toolkits.axisartist.grid_finder as GF

        self.refstd = refstd            # Reference standard deviation

        tr = PolarAxes.PolarTransform()

        # Correlation labels
        rlocs = NP.concatenate((NP.arange(10)/10.,[0.95,0.99]))
        tlocs = NP.arccos(rlocs)        # Conversion to polar angles
        gl1 = GF.FixedLocator(tlocs)    # Positions
        tf1 = GF.DictFormatter(dict(zip(tlocs, map(str,rlocs))))

        # Standard deviation axis extent
        self.smin = 0
        self.smax = 1.5*self.refstd

        ghelper = FA.GridHelperCurveLinear(tr,
                                           extremes=(0,NP.pi/2, # 1st quadrant
                                                     self.smin,self.smax),
                                           grid_locator1=gl1,
                                           tick_formatter1=tf1,
                                           )

        if fig is None:
            fig = PLT.figure()

        ax = FA.FloatingSubplot(fig, rect, grid_helper=ghelper)
        fig.add_subplot(ax)

        # Adjust axes
        ax.axis["top"].set_axis_direction("bottom")  # "Angle axis"
        ax.axis["top"].toggle(ticklabels=True, label=True)
        ax.axis["top"].major_ticklabels.set_axis_direction("top")
        ax.axis["top"].label.set_axis_direction("top")
        ax.axis["top"].label.set_text("Correlation")

        ax.axis["left"].set_axis_direction("bottom") # "X axis"
        ax.axis["left"].label.set_text("Standard deviation")

        ax.axis["right"].set_axis_direction("top")   # "Y axis"
        ax.axis["right"].toggle(ticklabels=True)
        ax.axis["right"].major_ticklabels.set_axis_direction("left")

        ax.axis["bottom"].set_visible(False)         # Useless
        
        # Contours along standard deviations
        ax.grid(False)

        self._ax = ax                   # Graphical axes
        self.ax = ax.get_aux_axes(tr)   # Polar coordinates

        # Add reference point and stddev contour
        print "Reference std:", self.refstd
        l, = self.ax.plot([0], self.refstd, 'k*',
                          ls='', ms=10, label=label)
        t = NP.linspace(0, NP.pi/2)
        r = NP.zeros_like(t) + self.refstd
        self.ax.plot(t,r, 'k--', label='_')

        # Collect sample points for latter use (e.g. legend)
        self.samplePoints = [l]

    def add_sample(self, stddev, corrcoef, *args, **kwargs):
        """Add sample (stddev,corrcoeff) to the Taylor diagram. args
        and kwargs are directly propagated to the Figure.plot
        command."""

        l, = self.ax.plot(NP.arccos(corrcoef), stddev,
                          *args, **kwargs) # (theta,radius)
        self.samplePoints.append(l)

        return l

    def add_contours(self, levels=5, **kwargs):
        """Add constant centered RMS difference contours."""

        rs,ts = NP.meshgrid(NP.linspace(self.smin,self.smax),
                            NP.linspace(0,NP.pi/2))
        # Compute centered RMS difference
        rms = NP.sqrt(self.refstd**2 + rs**2 - 2*self.refstd*rs*NP.cos(ts))
        
        contours = self.ax.contour(ts, rs, rms, levels, **kwargs)

        return contours


if __name__=='__main__':

    # Reference dataset
    x = NP.linspace(0,4*NP.pi,100)
    data = NP.sin(x)
    refstd = data.std(ddof=1)           # Reference standard deviation

    # Models
    m1 = data + 0.2*NP.random.randn(len(x))    # Model 1
    m2 = 0.8*data + .1*NP.random.randn(len(x)) # Model 2
    m3 = NP.sin(x-NP.pi/10)                    # Model 3

    # Compute stddev and correlation coefficient of models
    samples = NP.array([ [m.std(ddof=1), NP.corrcoef(data, m)[0,1]]
                         for m in (m1,m2,m3)])

    fig = PLT.figure(figsize=(8,8))
    
#    ax1 = fig.add_subplot(1,2,1, xlabel='X', ylabel='Y')
 
    # Taylor diagram
    dia = TaylorDiagram(refstd, fig=fig, rect=111, label="Reference")

    colors = PLT.matplotlib.cm.jet(NP.linspace(0,1,len(samples)))

#    ax1.plot(x,data,'ko', label='Data')
#    for i,m in enumerate([m1,m2,m3]):
#        ax1.plot(x,m, c=colors[i], label='Model %d' % (i+1))
#    ax1.legend(numpoints=1, prop=dict(size='small'), loc='best')

    # Add samples to Taylor diagram
    for i,(stddev,corrcoef) in enumerate(samples):
        dia.add_sample(stddev, corrcoef, marker='s', ls='', c=colors[i],
                       label="Model %d" % (i+1))

    # Add RMS contours, and label them
    contours = dia.add_contours(colors='0.5')
    PLT.clabel(contours, inline=1, fontsize=10)

    # Add a figure legend
    fig.legend(dia.samplePoints,
               [ p.get_label() for p in dia.samplePoints ],
               numpoints=1, prop=dict(size='small'), loc='upper right')


    PLT.show()



"""
ncexplorer.plotter.projection
-----------------------------

This module contains wrapper classes for some of the large set of projections
implemented in Matplotlib.pyplot.

@created: May 30, 2017
@author: neil
"""
from ncexplorer.const import LOC_UCLA as UCLA
from mpl_toolkits.basemap import Basemap


# Equidistant Cylindrical Projection.
# From Basemap Matplotlib Toolkit documentation: The simplest projection, just
# displays the world in latitude/longitude coordinates.
PROJ_EQUIDISTANT_CYLINDRICAL = 'cyl'

# Mercator
# From Basemap Matplotlib Toolkit documentation: A cylindrical, conformal
# projection. Very large distortion at high latitudes, cannot fully reach the
# polar regions.
PROJ_MERCATOR = 'merc'

# Lambert Azimuthal Equal Area.
# From Basemap Matplotlib Toolkit documentation: An equal-area projection. The
# green shapes drawn on the map are equal-area circles on the surface of the
# earth. Known as Tissot's indicatrix, they can be used to show the angular
# and areal distortion of a map projection. On a conformal projection, the
# shape of the circles is preserved, but the area is not. On a equal-area
# projection, the area is preserved but the shape is not.
PROJ_LAMBERT_AZIMUTHAL_EQUAL_AREA = 'laea'

# Lambert Conformal Projection.
# From Basemap Matplotlib Toolkit documentation: A conformal projection. The
# green shapes drawn on the map are equal-area circles on the surface of the
# earth. Known as Tissot's indicatrix, they can be used to show the angular
# and areal distortion of a map projection. On a conformal projection, the
# shape of the circles is preserved, but the area is not. On a equal-area
# projection, the area is preserved but the shape is not.
PROJ_LAMBERT_CONFORMAL = 'lcc'

# Orthographic Projection
# From Basemap Matplotlib Toolkit documentation: The orthographic projection
# displays the earth as a satellite (in an orbit infinitely high above the
# earth) would see it.
PROJ_ORTHOGRAPHIC = 'ortho'

# Double Orthographic Projection
# This is two orthographic projections side by side.  The left is as described
# above.  The right differs from the left in that the center of the right is
# the polar opposite point of the center of the left.
PROJ_ORTHOGRAPHIC_2 = 'ortho2'

SUPPORTED_PROJECTIONS = [PROJ_EQUIDISTANT_CYLINDRICAL,
                         PROJ_MERCATOR,
                         PROJ_ORTHOGRAPHIC,
                         PROJ_ORTHOGRAPHIC_2,
                         PROJ_LAMBERT_CONFORMAL,
                         PROJ_LAMBERT_AZIMUTHAL_EQUAL_AREA]

PROJECTION_DESCRIPTIONS = [
    {'projection': PROJ_EQUIDISTANT_CYLINDRICAL,
     'description': 'Equidistant Cylindrical'},
    {'projection': PROJ_MERCATOR, 'description': 'Mercator'},
    {'projection': PROJ_ORTHOGRAPHIC, 'description': "Orthoographic"},
    {'projection': PROJ_ORTHOGRAPHIC_2, 'description': "Double Orthoographic"},
    {'projection': PROJ_LAMBERT_CONFORMAL, 'description': "Lambert Conformal"},
    {'projection': PROJ_LAMBERT_AZIMUTHAL_EQUAL_AREA,
     'description': "Lambert Azimuthal Equal Area"}]


# Utility function to create a new class.  This function eliminates the need
# for other modules to have knowledge of class names.
def new_projector(projection):
    """Utility function for creating new instances of a projection class."""
    if projection == PROJ_EQUIDISTANT_CYLINDRICAL:
        return ProjEquidistantCylindrical()

    elif projection == PROJ_MERCATOR:
        return ProjMercator()

    elif projection == PROJ_ORTHOGRAPHIC:
        return ProjOrthographic()

#    elif projection == PROJ_ORTHOGRAPHIC_2:
#        return ProjOrthographic2()
    elif projection == PROJ_LAMBERT_CONFORMAL:
        return ProjLambertConformal()

    elif projection == PROJ_LAMBERT_AZIMUTHAL_EQUAL_AREA:
        return ProjLambertAzimuthal()


# Projection classes.
class Projector(object):
    """Base class for projections.
    
    The projection classes manage the canvas sizes and basemap initialization
    routines.
    """
    def __init__(self, projection):
        self.projection = projection
        self.projections = SUPPORTED_PROJECTIONS
    
    def __str__(self):
        return PROJECTION_DESCRIPTIONS[self.projection]
    
    # The method set_basemap is specific to the subclass and must be
    # implemented there.
    def set_basemap(self, ax):
        """Sets the projection and the region for the map."""
        self._map = self._set_basemap(ax)
        return self._map


class ProjOrthographic(Projector):
    """A Projector class for an orthographic projection.
    
    The orthographic projection requires a center.  The campus of UCLA is the
    default center.
    """
    def __init__(self, center=None):
        Projector.__init__(self, PROJ_ORTHOGRAPHIC)
        
        if center is None:
            self.center = UCLA
        else:
            self.center = center

    def _set_basemap(self, ax):
        mapobj = Basemap(ax=ax,
                      projection=self.projection,
                      lat_0=self.center[0],
                      lon_0=self.center[1])
        return mapobj


# This class
class ProjLambertConformal(Projector):
    """A Projector class for the Lambert conformal projection.
    
    The default size is 12,000 km wide by 9,000 km high and an ellipsoid of
    approximately 6378 km by 6357 km.
    """
    def __init__(self):
        Projector.__init__(self, PROJ_LAMBERT_CONFORMAL)

        # FIXME: Calculate these parameters from a simpler request such as a
        # center and a size.
        self.params = {
            'width': 12000000,
            'height': 9000000,
            'rsphere': (6378137.00, 6356752.3142),
            'resolution': 'l',
            'area_thresh': 1000.0,
            'lat_1': 45.0,
            'lat_2': 55.0,
            'lat_0': 50.0,
            'lon_0': -107.0}

    def _set_basemap(self, ax):
        mapobj = Basemap(ax=ax,
                         projection=self.projection,
                         width=self.params['width'],
                         height=self.params['height'],
                         rsphere=self.params['rsphere'],
                         resolution=self.params['resolution'],
                         area_thresh=self.params['area_thresh'],
                         lat_1=self.params['lat_1'],
                         lat_2=self.params['lat_2'],
                         lat_0=self.params['lat_0'],
                         lon_0=self.params['lon_0'])
        return mapobj


class ProjLambertAzimuthal(Projector):
    """A Projector class for the Lamber azimuthal projection."""
    def __init__(self):
        Projector.__init__(self, PROJ_LAMBERT_AZIMUTHAL_EQUAL_AREA)

        # FIXME: Calculate these parameters from a simpler request such as a
        # center and a size.
        self.params = {
            'width': 12000000,
            'height': 8000000,
            'resolution': 'l',
            'lat_ts': 50.0,
            'lat_0': 50.0,
            'lon_0': -107.0}

    def _set_basemap(self, ax):
        mapobj = Basemap(ax=ax,
                         projection=self.projection,
                         width=self.params['width'],
                         height=self.params['height'],
                         resolution=self.params['resolution'],
                         lat_1=self.params['lat_ts'],
                         lat_0=self.params['lat_0'],
                         lon_0=self.params['lon_0'])
        return mapobj


# This class is the parent class for map projects that fit into a rectangle.
# The defining property is the lower-left corner and the upper-right corner.
class ProjectorFlat(Projector):
    
    def __init__(self, projection):
        Projector.__init__(self, projection)

        # FIXME: It would be nice to go from 180 to 180, but that either
        # produces a div/0 exception, or (-180, 180) paints only half the map.
        self.corners = {
            'llcrnrlat': -90.0,
            'llcrnrlon': 0,
            'urcrnrlat': 90.0,
            'urcrnrlon': 360}
        
    def _set_basemap(self, ax):
        mapobj = Basemap(ax=ax,
                      projection=self.projection,
                      llcrnrlat=self.corners['llcrnrlat'],
                      llcrnrlon=self.corners['llcrnrlon'],
                      urcrnrlat=self.corners['urcrnrlat'],
                      urcrnrlon=self.corners['urcrnrlon'])
        return mapobj


class ProjEquidistantCylindrical(ProjectorFlat):
    """A Projector class for the equidistant cylindrical projection.
    
    The boundaries of this projection start at -90 deg latitude to +90 deg
    latitude, and 0 deg longitude to 360 deg longitude.
    """
    def __init__(self):
        ProjectorFlat.__init__(self, PROJ_EQUIDISTANT_CYLINDRICAL)

        
class ProjMercator(ProjectorFlat):
    """A Projector class for the Mercator projection.
    
    The boundaries of this projection are -80 deg latitude to +80 deg
    latitude and 0 deg longitude to 360 deg longitude.
    """
    def __init__(self):
        ProjectorFlat.__init__(self, PROJ_MERCATOR)
        self.corners = {
            'llcrnrlat': -80.0,
            'llcrnrlon': 0,
            'urcrnrlat': 80.0,
            'urcrnrlon': 360}



'''
Created on Apr 24, 2017

@author: neil
'''
# Locations
LOC_JUNEAU = [58.3019, -134.4197]
LOC_MANAUS = [-3.117034, -60.025780]
LOC_UCLA = [34.07, -118.45] 

# Map colorings.
COLOR_COASTLINES = '#666666'
COLOR_CONTINENTS = '#cccccc'


'''
Created on Dec 22, 2016

@author: neil
'''
import ConfigParser, os

config = ConfigParser.ConfigParser()

# FIX ME: Change the default location
default_loc = '/home/neil/workspace/proj-eclipse/ncexplorer/etc/ncexplorer.cfg'
config.read([default_loc, os.path.expanduser('~/.ncexplorer.cfg')])

# ESGF Repository
CFG_ESGF_NODE = config.get('ESGF', 'esgf_node')
CFG_ESGF_SEARCH_NODE = config.get('ESGF', 'esgf_search_node')
CFG_ESGF_OPENID_NODE = config.get('ESGF', 'esgf_openid_node')

# NASA Earthdata Repository
URS_SERVER = config.get('NASA Earthdata', 'urs_server')
URS_DIRECTORY = config.get('NASA Earthdata', 'urs_directory')

# Local repository directories
repodirs = config.items('Directory Repositories')

# Matplotlib backends
MPL_BACKEND_REQUIREMENT = config.get('Matplotlib', 'mpl_backend_requirement')

# Username and password to avoid logging in multiple times.
TRIVIAL_USERNAME = config.get('Authentication', 'username')
TRIVIAL_PASSWORD = config.get('Authentication', 'password')

# Package the repositories up for consumption by the application.
# TODO: Get a dynamic list of repository servers from the config file.
repositories = []
repositories.append({
    'type': 'esgf',
    'parameters': {
        'node': CFG_ESGF_NODE,
        'search_node': CFG_ESGF_SEARCH_NODE,
        'openid_node': CFG_ESGF_OPENID_NODE
        }
    })
repositories.append({
    'type': 'urs',
    'parameters': {
        'server': URS_SERVER,
        'directory': URS_DIRECTORY
        }
    })
for key, path in repodirs:
    repositories.append({
        'type': 'local',
        'parameters': {
            'path': path
            }
        })



"""
Created on Mar 28, 2017

@author: neil

These plotters use the TkAGG interactive backend.  They are intended for a
console application or the desktop.
"""
from mpl_toolkits.basemap import Basemap
from mpl_toolkits.basemap import addcyclic
import numpy as np
#import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from ncexplorer.plotter.plotter import MapPlotter
# from pygments.lexers.basic import BlitzMaxLexer

        
# This class servers the console app.
class BasemapPlotter(MapPlotter):
    
#    def __init__(self, **kwargs):
#
#        MapPlotter.__init__(self, **kwargs)

    # this method requires that the pyplot axes has been already created.
    def _set_basemap(self):
        self._map = self._projector.set_basemap(self._ax)

    def clear(self):
        
        self._ax.cla()
        self._map.drawmapboundary()
        self._map.drawcoastlines(linewidth=0.5, color=self._color_coastlines)
        self._map.fillcontinents(color=self._color_continents, alpha=0.5)

    # Plot a point.
    def plotpoints(self, points, color=None, pointsize=None):
        
        lonlats = self.points_to_lonlats(points)
        if color is None:
            color = '#0000FF'
        if pointsize is None:
            pointsize = 4
        
        x, y = self._projector._map(lonlats[0], lonlats[1])
        self._map.plot(x, y, 'o', color=color, markersize=pointsize)

    def _draw(self, **kwargs):
        
        # It's possible to call draw() without the dataset being set as an
        # attribute of the parent class.  so handle that case with an
        # informative error.
        try:
            var = self.dataset.data_vars['var']
        except AttributeError:
            msg = "Has the dataset been set?"
            raise AttributeError(msg)
            

        # TODO: The (x,y)-coordinate grid does not need to be recalculated
        # every time.  Check for compatibility, and recalculate only when
        # necessary.
        lat = var['lat']
        lon = var['lon']
        
        # Make the data wrap around the globe, instead of leaving a slice of
        # empty space between the last longitude and the first.
        #
        # Sometimes this works, sometimes it fails to broadcast the input
        # array.  In this case, we'll take the plot with a slice missing.
        try:
            data, cyclic_lon = addcyclic(var, lon)
        except ValueError:
            data = var
            cyclic_lon = lon
        except IndexError:
            # These seems to have started with a new version of a library.
            # FIXME: Find out why this exception occurs.
            data = var
            cyclic_lon = lon
        
        lons, lats = np.meshgrid(cyclic_lon, lat)
        xl, yl = self._map(lons, lats)
        cs = self._map.contourf(xl, yl, data, levels=self.contour_levels, colors=self.contour_colors)#, alpha=0.5)

        # The _showgrid attribute means that the grid (lat, lon) should be
        # displayed on the map.
        if self._showgrid:
            self._map.drawparallels(lat.values)
            self._map.drawmeridians(lon.values)

        # FIXME: Once the reason for this causing an error in jupyter notebooks
        # is discovered, this test can be removed.
        if self._colorbar and self._canvas.colorbar_ok():
            self._map.colorbar(cs)


# This class serves the Tk desktop GUI application.
class TkPlotter(BasemapPlotter):
    
    # Only the __init__ method needs to be implemented.  The Tk frame manages
    # the canvas on which the map is drawn.  The remaining methods are
    # indifferent to how the canvas is managed.
    def __init__(self, dpi, width, height, center=None):
        """Sets the Tk canvas as the figure and axes."""

        self.figure = Figure(dpi=dpi, figsize=(width,height))
        self.axes = self.figure.add_subplot(111)
        
        # Need the parent to initialize.
        BasemapPlotter.__init__(self, center)

    # This is the call that's specific to this class.
    def _set_basemap(self):
        self._map = Basemap(ax=self.axes,
                            projection='ortho',
                            lat_0=self._center[0],
                            lon_0=self._center[1])



"""First attempt at setup.py"""

from setuptools import setup, find_packages


setup (
	name='ncexplorer',
	version='0.0.1.dev5',
	description='Climate data analysis utility.',
	long_description='Climate data analysis utility.',
	url='https://github.com/godfrey4000/ncexplorer',
	license='MIT',
	classifiers=[
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Science/Research',
		'Topic :: Scientific/Engineering :: Atmospheric Science',
		'License :: OSI Approved :: MIT License',
		'Programming Language :: Python :: 2.7',
	],
	keywords='climate netcdf analysis',

	# The packages
	packages=find_packages(exclude=['docs', 'etc', 'ncexplorer/test']),
#	install_requires=['xarray', 'esgf-pyclient'],
)




"""
The frame.base module
---------------------

This module is where the frame interface definition resides.  The interface
consists of designated private methods of the BaseFrame class.  A new user
interface can be implemented by subclassing the BaseFrame class and
implementing these methods.

.. parsed-literal::
   @created: Mar 11, 2017
   @author: neil
"""
from ncexplorer.app import Application
from ncexplorer.repository import NCXESGF
#from ncexplorer.plotter.canvas import PlottingCanvas
from ncexplorer.plotter.canvas import NotebookCanvas


# Extract human understandable names from the variable property of datasets.
def parse_variable_names(var):
    """Parse the variable names.
    
    Some variables don't have original names, and some don't have long names.
    Show a variable name, a long name and the coordinates, choosing
    attributes in this order:

        var name:  original_name, vkey
        long name: long_name, standard_name, vkey
    """
    if 'long_name' in var.attrs.keys():
        longname = var.attrs['long_name']
    elif 'standard_name' in var.attrs.keys():
        longname = var.attrs['standard_name']
    else:
        longname = var.name

    return {'name': var.name, 'longname': longname}


# Since there are many processes that take a while (e.g., searching a
# repository of NetCDF files, downloading them, cleaning and regridding
# data) the full stack makes extensive use of progress bars.  Otherwise,
# the application appears unresponsive.
#
# The presence of this base class establishes the interface the application
# stack requires and a default implementation.
class BaseProgressBar(object):
    """Implements the progress bar interface."""
    def update(self, msg=""):
        """Updates the progress bar."""
        pass
    def start(self, total_steps):
        """Resets the progress bar.
        
        total_steps:  The total number of steps to completion.  Each time the
            method update() is called, the step increases by one.  The percent
            complete is the number of updates()/total_steps.
        """
        pass
    def close(self):
        """This does not do anything."""
        pass

# A dictionary that will give a dataset as an attribute.
# FIXME: Consider subclassing of collections.abc or UserDict
class DatasetDictionary(dict):
    
    @property
    def dataset(self):
        if 'dataset' in self.keys():
            return self['dataset']
    

class BaseFrame(object):
    """The parent class of the family of frame classes.
    
    title:
        The title for the application.  The title is displayed in logs and
        on the window title bar, among other places.
    """
    def __init__(self, title):
        # These methods set the UI objects that the application will call.
        self._progressbar = self._set_progressbar()
        
        # This method sets the application subclass.  The Application object
        # is the main controller -- the heart of the application.
        self._app = self._set_application(title)
        
        # The frame subclass must create the canvas.  The details of the
        # canvas implementation could relate to a web browser, a desktop
        # X window, etc.
        self._canvas = self._new_canvas()

    # These methods constitute the interface.
    def _set_application(self, title):
        """This method must instantiate and return an application instance.
        """
        pass
    def _set_progressbar(self):
        """This method must instantiate and return a progress bar instance.
        """
        pass
    def _plotter(self, **kwargs):
        """FIX ME: this method should instantiate and return an instance of
        a plotting canvas.
        """
        pass
    def _display_matches(self, matches):
        """This method handles displaying the NetCDF files that match a search
        request.
        """
        pass
    def _display_variables(self, payload):
        """This method handles displaying the variables contained in a
        collection of NetCDF files.
        """
        pass
    def get_login_creds(self):
        """This method prompts the user for username and password to be used
        to authenticate to a repository if it is required.
        """
        pass

    def _new_canvas(self):
#        canvas = PlottingCanvas(figsize=(6,6))
        canvas = NotebookCanvas(figsize=(6,6))
        return canvas
        
    def plotter(self, **kwargs):
        """Returns a new instance of the plotter object.
        
        Parameters
        ----------
        plottype : string
            The plottype can either be 'map', or 'scatter'.
        """
        return self._plotter(**kwargs)

    # Useful attributes, particularly when using these objects from the python
    # interpreter.
    # FIX ME: These should python objects instead of console output.
    @property
    def usernames(self):
        """A list of the user names associated with the repositories."""
        return self._app.list_usernames()

    @property
    def variables(self):
        """A list of all the variables known to the application."""
        self.display_variables(self._app.datasets)

    @property
    def datasets(self):
        """A list of the datasets known to the application."""
        retval = []
        for index, ds in self._app.datasets.iteritems():
#            dsdictionary = {}
            dsdictionary = DatasetDictionary()
            dsdictionary['ID'] = index
            dsdictionary['dataset'] = ds
            if not callable(ds):
                if 'institute_id' in ds:
                    dsdictionary['institute'] = ds.institute_id
                if 'institute_id' in ds:
                    dsdictionary['model'] = ds.model_id
                if 'institute_id' in ds:
                    dsdictionary['experiment'] = ds.model_id
            retval.append(dsdictionary)

        return retval 

    # These methods are application methods, exposed here for convenience.
    def variable(self, varspec):
        """Returns a variables.
        
        Takes a specification of the form dataset_index:variable_name and
        returns the xarray DataArray object representing that variable.
        """
        return self._app.variable(varspec)

    def save(self, request):
        """Saves the application's save method."""
        self._app.save(request)

    # This is a higher level workflow.  It accepts a search string from the
    # user and calls the app's search method with it.  The user then presumably
    # selects which of these should be downloaded.  That's done in the bind()
    # method.
    def search(self, **kwargs):
        """Performs a search across all repositories.
        
        This method calls the applications search method.  See the
        application's description for more information.  The application will
        call the display_matches method to let the frame handle displaying
        the results to the user.
        """
        self._app.search(**kwargs)
        matches = self._app.search_results()
        self._display_matches(matches)
        return matches

    def bind(self, matchlist):
        """Builds variables from the selected files.
        
        Takes a subset of the list of URLs that matched the search (matchlist)
        and calls the apps bind() method.  Details of the contents of the files
        and the variables are pushed to the UI.
        
        The matchlist is a list of tuples which have the form (repo, file).
        The file can be either the integer or the filename.
        """        # Enforce type 
        # The app will push the data to the frame in the display_variables()
        # method.  For now, all file are selected.
        self._app.bind_data(matchlist)

    # Methods to support the application object.
    def progressbar(self, dummy):
        """Returns a progress bar."""
        return self._progressbar

#    def get_login_creds(self):
#        '''Retrieve a username and password from the command line.'''
#        print "Login required"
#        username = raw_input('Username: ')
#        password = getpass.getpass('Password: ')
#        ret = {'username': username, 'password': password}
#        return ret
    
    def get_search_params(self):
        """Prompts user for a search parameter string
        
        Returns the search parameter string.
        """
        print "Enter search parameters like this: variable=ta, institute=NCAR"
        param_string = raw_input('Keyword string: ')
        return param_string

#    def display_repo_matches(self):
#        '''
#        List all the URLS found from a search in the given repository.  The
#        output is short to accommodate Jupyter notebooks.
#        '''variables
#        matches = self._app.search_results(filenameonly=True)
#        return self._display_matches(matches)
    
    def display_variables(self, datasets):
        """
        Package all datasets and their properties into a list suitable
        for display in a tree view.  Each dataset itself, contains a set of
        variables which, depending the the capabilities of the frame subclass,
        would be visible when expanding a specified dataset.
        """
        payload = []
        for index, ds in datasets.iteritems():
            # FIX ME: Preferably the repository should prepare the list of
            # variables for the tree view.  That's where knowledge about the
            # datasets resides.  The tree view just displays the list, and
            # lets the user select a subset.
            # 
            # For now, build a tuple of the form:
            #    (dataset, variable)
            # and use that as the tag.
            #
            # The dataset (ds) could be a generator.
            if not callable(ds) and (
                'institute_id' in ds and
                'model_id' in ds and
                'experiment_id' in ds):
                dsdictionary = {
                    'ID': index,
                    'Institute': ds.institute_id,
                    'Model': ds.model_id,
                    'Experiment': ds.experiment_id}
            else:
                dsdictionary = {
                    'ID': index,
                    'Institute': 'unknown',
                    'Model': 'unknown',
                    'Experiment': 'unknown'}
            
            # FIX ME: The determination of the names belongs in the
            # application.
            vars = []
            if callable(ds):
                names = {'name': 'unknown', 'longname': 'unknown'}
                vars.append(names)
            else:
                for vkey in ds.data_vars:
                    var = ds[vkey]
                    names = parse_variable_names(var)
                    vars.append(names)
            dsdictionary['Variables'] = vars

            payload.append(dsdictionary)
        
        # Leave it to the subclass to display the datasets and their variables
        # to the user in the most appropriate way.
        self._display_variables(payload)

    # Utility functions.  These just call the corresponding app method.
    def regrid(self, var, likevar=None):
        return self._app.regrid(var, likevar)

    # This method must be overridden.
    def mainloop(self):
        raise NotImplemented(
            "The mainloop method of the BaseFrame cannot be called " +
            "directly.  Derive a class from BaseFrame and implement " +
            "that mainloop method.")



'''
Created on Mar 11, 2017

@author: neil
'''
from ncexplorer.frame.base import BaseFrame
from ncexplorer.plotter import Plotter
import mpld3

class D3Plotter(Plotter):
    '''
    A subclass of Plotter that uses mpld3 to render an HTML version of the
    plot for display in a browser/JS based GUI.
    '''
    def __init__(self, outfn):
        
        self._outfn = outfn

        # The base object uses the default backing, I believe.
        Plotter.__init__(self, figure=True)

    # This function is overriden so that it can push the output to the
    # javascript client.
    def plotdata(self, var, time_=0, plev=0):
        # The parent's draw() method creates the plot.
        try:
            Plotter.draw(self, variable=var, time_=time_, plev=plev)
        except IOError as (errno, strerror):
            msg = "I/O error({0}): {1}".format(errno, strerror)
            self._outfn('error', msg)
            raise
        except:
            msg = "Unexpected error."
            self._outfn('app-error', msg)
            raise
            
        # Then creates and HTML version of the plot and pushes it to the
        # client.
        fightml = self._html()
        self._outfn('plot', {'html': fightml})

    def _html(self):
        fightml = mpld3.fig_to_html(self._figure)
        return fightml

#class DSPlotter(Plotter):
#    def getfig(self):
#        self.get_vars()
#        fig, ax = plt.subplots()
#        plotter = Plotter(axes=ax)
#        plotter.draw(axes=ax, var=self._variable)
#        fightml = mpld3.fig_to_html(fig)
#        return fightml


class DummyProgressbar(object):
    '''A progress bar stub that only holds the data measuring progress.'''
    def __init__(self, outfn):
        self._msg = ""
        self._total_steps = None
        self._step = 0
        self._status = 'not started'

        # The output function.
        self._outfn = outfn

    def update(self, msg=""):
        self._step += 1
        self._msg = msg
        self._send_update()
        
    def start(self, total_steps):
        self._total_steps = total_steps
        self._status = 'in progress'
        
    def close(self):
        self._status = 'completed'
        self._send_update()

    def status(self):
        '''Provides a dictionary with details of the status of a job.'''
        frac = float(self._step)/float(self._total_steps)
        percent = ("{0:." + str(2) + "f}").format(100 * frac)
        ret = {
            'status': self._status,
            'step': self._step,
            'total_steps': self._total_steps,
            'percent': percent,
            'msg': self._msg
        }
        return ret
    
    def _send_update(self):
        data = self.status()
        self._outfn(data)

class FlaskGatewayFrame(BaseFrame):
    
    def __init__(self, title, clientout):

        # Need to override the __init__ to setup the output channel to the
        # javascript client.
        self._clientout = clientout
        
        # The progress bars use the clientout.
        self._search_progress_bar = DummyProgressbar(self._search_out)
        
        # The parent's __init__ still required.
        BaseFrame.__init__(self, title)

    # Output channels.  These are just event/message pairs.
    def _search_out(self, msg):
        self._clientout('pbsearch', msg)

#    def _display_url_out(self, msg):
#        self._clientout('search-result', msg)

    # Methods required to be implemented.
    def _display_matches(self, matches):
        self._clientout('search-result', matches)

    def _set_progressbar(self):
        return self._search_progress_bar

    def _display_variables(self, payload):
        self._clientout('variables', payload)

    def _set_plotter(self):
        plotter = D3Plotter(self._clientout)
        return plotter

    # Nothing implemented here.  The main loop in this case is the web server
    # provided by Flask.
    def mainloop(self):
        raise NotImplemented("No mainloop implemented from the flask " +
                             "gateway.  The flask web server provides " +
                             "this function.")



"""
The plotter module implements the base class Plotter.  mapPlotter is a subclass
specialized for maps (displaying data on a map of the surface of the Earth).
ScatterPlotter is a subclass for creating scatter plots.

@created: Mar 28, 2017
@author: neil
"""
from ncexplorer.plotter.projection import SUPPORTED_PROJECTIONS
from ncexplorer.plotter.projection import PROJECTION_DESCRIPTIONS
from ncexplorer.plotter.projection import PROJ_ORTHOGRAPHIC
from ncexplorer.plotter.projection import new_projector
from ncexplorer.const import COLOR_CONTINENTS, COLOR_COASTLINES 
from matplotlib.ticker import FixedLocator
import numpy as np
import xarray as xr


class Plotter(object):
    """The Plotter interface.
    
    There are two types of plots: scatter plots and maps.  The base class for
    scatter plots is ScatterPlotter.  The base class for maps is MapPlotter.
    
    Any Plotter class has a dataset attribute which is an xarray instance of
    a DataArray or a Dataset containing the data to be plotted.
    """
    # This implementation requires that subclasses of this class implement a
    # setter method for the dataset attribute.
    def __init__(self, **kwargs):
        
        if 'dataset' in kwargs and kwargs['dataset'] is not None:
            self.dataset = kwargs['dataset']
        
        # The subclass must set the charttype.
        self.charttype = 'None'
        self._title = None

    def set_defaults(self):
        pass

    def set_canvas(self, canvas):
        """Sets the canvas and plotting area for the plot.
        
        The canvas is the window or region in the user interface where all the
        plots are displayed.  Each individual plot is a subsection or subregion
        of this area.  In addition to setting the canvas, this method calls
        the canvas add_map method to create specific plotting area.
        """
        self._canvas = canvas

        # Schematically, the canvas is the region on which the scatter plots
        # and maps are drawn.  More concretely, it is the Matplotlib.pyplot's
        # figure instance.  The add_plot method creates a rectangle on the
        # figure for a particular plot.
        ax = self._add_plot()
        return ax

    def draw(self):
        """Render the data as contour levels on a map."""
        self._draw()
#        self._canvas.show()

    def clear(self):
        """Draws the map with continents, but no data.
        
        This method effectively clears the map."""
        # Maybe this will clear the figure.
        self._canvas.clear()
        
        self._clear()
#        self._canvas.show()
    
    def update(self):
        """Calls clear() and then draw() to update the plot."""
        self.clear()
        self.draw()

    def savefig(self, filename):
        """Save the figure to disk as a PNG file.
        
        This calls the matplotlib savefig() method.
        """
        fig = self._canvas._figure
        fig.savefig(filename)

    @property
    def projections(self):
        """Displays a list of supported projections.
        
        Map type plotters will return a list of supported projections.  Other
        types where it's not appropriate will return the short message
        explaining that the plotter type doesn't support projections.
        """
        return "Chart type {0} does not support projections.".format(
            self.charttype)

    @property
    def charttypes(self):
        return ['Skatter', 'Map']


    # The utility methods for printing the object.  These methods are useful
    # when debugging.  The debugging window in eclipse calls the __repr__ and
    # __str__ methods.
    def _pretty(self):
        # The subclass implements this.
        pass

    def __str__(self):
        return unicode(self).encode('utf-8')
    
    def __repr__(self):
        return unicode(self)

    def __unicode__(self):
        pretty = self._pretty()
        return u'\n'.join(pretty)

    @property
    def dataset(self):
        """The object containing the plotted data.
        
        The object must be an xarray DataArray or Dataset, and conform to the
        following conventions:

            The object can be an xarray DataArray object, and it must have the
            appropriate dimensions defined.
            
            The object can be an xarray Dataset object with a single variable,
            and that variable must have the appropriate dimensions.
            
            The object can be an xarray Dataset object with more than one
            variable, but one of the variables must be 'var'.  That is the
            variable that is plotted.
        
        In any of these cases, object is converted to an xarray Dataset that
        conforms to the last condition.
        """
        return self._dataset

    @dataset.setter
    def dataset(self, dataobj):

        # xarray DataArrays can be plotted if they have 'lat' and 'lon'
        # dimensions
        if type(dataobj) is xr.DataArray:
            newds = dataobj.to_dataset(name='var')
            var = newds.data_vars['var']
            if self._validate_dimensions(var):
                self._dataset = newds
            
        # If the first parameter isn't an xarray DataArray, then it must be an
        # xarray Dataset.
        elif type(dataobj) is not xr.Dataset:
            msg = ("The first parameter must be a xarray DataArray or " +
                   "Dataset object.")
            raise TypeError(msg)

        # If the xarray Dataset has only one variable, presume the intention
        # is to plot that variable.
        elif len(dataobj.data_vars) == 1:
            
            # That only only one entry exists in the dictionary data_vars has 
            # been establish.  The key is unknown, but the following access in
            # this case is safe.
            var = dataobj.data_vars.values()[0]
            if self._validate_dimensions(var):
                newds = var.to_dataset(name='var')
                self._dataset = newds
            
        # The dataset has more than one variable, then it must have a avariable
        # 'var', and that is the one to be plotted.  The legitimate instance of
        # this case is when data plots are saved in Datasets.
        elif not 'var' in dataobj.data_vars:
            msg = ("The variable var not found in xarray Dataset containing "
                   "more than one variable.")
            raise TypeError(msg)

        else:
            # It is now established that dataobj is an xarray Dataset with
            # exactly one variable named 'var'.
            var = dataobj.data_vars['var']
            
            # The dataset must have a 'lat' and a 'lon' dimension.
            if self._validate_dimensions(var):
                self._dataset = dataobj
            
        return


# ScatterPloting object
class ScatterPlotter(Plotter):
    """The ScatterPlotter interface.
    
    Attributes
    ----------
        dataset: (object) An xarray DataArray or Dataset containing the data
        to be plotted.
    """
    def __init__(self, dataset=None):
        
        # The plotter subclasses should call the parent, first thing.
        Plotter.__init__(self, dataset=dataset)
        self.charttype = 'Scatter'
        
    def _add_plot(self):
        self._ax = self._canvas.add_map()
        return self._ax
#        
#        self.figure = plt.figure(figsize=(8,2))
#        self.ax = self.figure.add_axes([0.1, 0.1, 0.88, 0.88])
#        self.ax.minorticks_on()
#        pass

    # The ScatterPlotter needs the time dimension.
    def _validate_dimensions(self, var):
        if not ('time' in var.dims):
            msg = "The dimension \'time\' must be defined for a scatter plot."
            raise TypeError(msg)
        return True

    def _clear(self):
        self.ax.cla()
        
    def xticks(self, ticklist):
        xlocator = FixedLocator(ticklist)
        self._ax.xaxis.set_major_locator(xlocator)
        self._ax.set_xlim([min(ticklist), max(ticklist)])

    def yticks(self, ticklist):
        ylocator = FixedLocator(ticklist)
        self._ax.yaxis.set_major_locator(ylocator)
        self._ax.set_ylim([min(ticklist), max(ticklist)])

    def addline(self, time_, var, linetype='thick'):
        if linetype == 'thin':
            linewidth = 0.25
            color = '#bce8ce'
        else:
            linewidth = 0.50
            color = '#053619'

        retline = self._ax.plot(time_, var, '-', color=color, linewidth=linewidth)
        return retline

    def addsquare(self, x, y, **kwargs):
        self._addpoint(x, y, 's', **kwargs)

    def addcircle(self, x, y, **kwargs):
        self._addpoint(x, y, 'o', **kwargs)
        
    def _addpoint(self, x, y, shape, **kwargs):
        
        if 'color' not in kwargs:
            color='#053619'
        else:
            color = kwargs['color']

        self._ax.plot(x, y, 's', color=color, markersize=8.0)

    # Return a list of properties for the __str__ family of functions.
    def _pretty(self):
        pretty = []
        pretty.append("Type: scatter")
        return pretty
    

# The base class defines the interface.
class MapPlotter(Plotter):
    """The MapPlotter interface.
    
    Attributes
    ----------
        center: (tuple) The latitude and longitude of the center of the map.
        
        contour_colors: (list of CSS color specifications: '#rrggbb') The
        colors used for the contours.
        
        contour_levels: (list) The break points for the contour levels.
        
        dataset: (object) An xarray DataArray or Dataset containing the data
        to be plotted.
        
        projection: (string) The map projection.
    
    Methods
    -------
        emptymap(center=None): Draws an map with just the continent outlines.
        This method effectively clears the map.
        
        savefig(filespec): Saves the plot as a PNG file to the location
        specified with filespec.
    """
    def __init__(self, dataset=None):
        
        # The plotter subclasses should call the parent, first thing.
        Plotter.__init__(self, dataset=dataset)
        self.charttype = 'Map'
        
        # NOTE: The properties being set here are actually being set by the
        # setter methods defined below.
        self._projector = None
        self._showgrid = False
        
        self._color_continents = COLOR_CONTINENTS
        self._color_coastlines = COLOR_COASTLINES
        
        # The colors and contour levels are set to none, so by default, they
        # will be chosen automatically by the basemap contourf method.
        # chosen automatically.
        self._cs_levels = None
        self._cs_colors = None
        self._colorbar = True

    def set_defaults(self):
        self.projection = PROJ_ORTHOGRAPHIC

    def _add_plot(self):
        self._ax = self._canvas.add_map()
        return self._ax


    @property
    def center(self):
        """The latitude and longitude of the center of the map.
        
        A tuple (latitude, longitude) that is used by the plotting library.
        The map that is drawn is centered on this latitude and longitude.
        """
        if self._projector is None:
            msg = "Projection is undefined."
            raise TypeError(msg)
        
        return self._projector.center

    @center.setter
    def center(self, center):
        # FIX ME: Add error handling error.  Raise and exception if the center
        # specified does not conform.
        if self._projector is None:
            msg = "Projection is undefined.  Cannot set center location."
            raise TypeError(msg)
        self._projector.center = center
        self._set_basemap()

    @property
    def colorbar(self):
        return self._colorbar

    @colorbar.setter
    def colorbar(self, colorbar_on):
        self._colorbar = colorbar_on

    @property
    def contour_colors(self):
        """The user-defined contour colors.
        
        If the contour colors have not been explicitly set by the user, then
        the matplotlib library will use a default color scheme.
        """
        return self._cs_colors

    @contour_colors.setter
    def contour_colors(self, colors):
        self._cs_colors = colors

    @property
    def contour_levels(self):
        """The user-defined contour levels.
        
        If the contour levels have not been explicitly set by the user, then
        the matplotlib library will calculate contour levels from the data
        being plotted.
        """
        return self._cs_levels

    @contour_levels.setter
    def contour_levels(self, levels):
        self._cs_levels = levels

    # Subclass must validate the dataset dimensions.
    def _validate_dimensions(self, var):
        # MapPlotters require latitude and longitude.
        if not ('lat' in var.dims and 'lon' in var.dims):
            msg = ("Map plots require \'lat\' (latitude) and \'lon\' " +
                   "(longitude).")
            raise TypeError(msg)
        return True
    
    @property
    def grid(self):
        """display the lines of latitude and longitude.
        
        The lines displayed correspond to the values for latitude and
        longitude in the dataset    @property
    def charttypes(self):
        return ['']

.  The grid lines are not displayed by default.
        """
        return self._showgrid

    @grid.setter
    def grid(self, on_boolean):
        if on_boolean not in (True, False):
            msg = "The grid attribute must be True or False."
            raise TypeError(msg)
        self._showgrid = on_boolean

    @property
    def projection(self):
        """The map projection.
        
        Projections supported are:
            Lambert Azimuthal Equal Area: 'laea'
            Lambert Azimuthal Conformal: 'lcc'
            Orthographic: 'ortho'
        
        The orthographic optionally supports two projections side-by-side.  The
        right-hand map is centered on the point diametrically opposite the
        specified center.
        """
        return self._projector.projection

    @projection.setter
    def projection(self, projection):
        # If the projection is already set to what has been requested, there
        # is no need to do anything at all.
        if (self._projector is not None and
            self._projector.projection == projection):
            return

        if projection not in SUPPORTED_PROJECTIONS:
            msg = ("The specified projection \'{0}\' is not " +
                   "supported.").format(projection)
            raise TypeError(msg)
        self._projector = new_projector(projection)
        self._set_basemap()

    @property
    def projections(self):
        """Return a list of supported projections"""
        return PROJECTION_DESCRIPTIONS

    def points_to_lonlats(self, points):
        """Converts a tuplelike set of points to longitude and latitude."""
        
        # Points are defined latitude first, longitude second.
        # Points can be a single tuple.
        if type(points) is tuple:
            lats = points[0]
            lons = points[1]
            lonlats = np.meshgrid(lons, lats)
            return lonlats

        # FIX ME:  Through an exception here.
        return None

    def _pretty(self):
        pretty = []
        pretty.append("Type: map")
        pretty.append("Projection: {0}".format(self.projection))
        pretty.append("Center: {0}".format(self.center))
        return pretty



'''
Created on May 29, 2017

@author: neil
'''
import matplotlib.pyplot as plt


class PlottingCanvas(object):
    
    def __init__(self, figsize):
        self._figure = plt.figure(figsize=figsize)
        self._figsize = figsize
        self._axes = []

    # FIXME:  This is a real kludge.  In jupyter notebooks, the colorbar
    # provokes a IndexException.  This is a temporary work-around until the
    # cause of this can be found and fixed.
    def colorbar_ok(self):
        return True

    def add_map(self):
        # FIXME: Manage the canvas real estate.
        ax = self._figure.add_subplot(111)
        self._axes.append(ax)
        return ax

    def clear(self):
        self._figure.clf()

    def show(self):
#        self._figure.show()
        pass

    def __unicode__(self):
        pretty = []
        pretty.append("Pyplot instance: %s" % plt)
        pretty.append("Figure instance: %s" % self._figure)
        pretty.append("Number of subplots: %s" % len(self._axes))
        return u'\n'.join(pretty)
    
    def __repr__(self):
        return unicode(self)
    
    def __str__(self):
        return unicode(self).encode('utf-8')


class NotebookCanvas(object):
    
    def __init__(self, figsize):
        self._figure = plt.figure(figsize=figsize)

    def colorbar_ok(self):
        return True

    def add_map(self):
        ax = self._figure.add_subplot(111)
        return ax

    def clear(self):
        pass

    def show(self):
#        plt.show(block=False)
        plt.ion()



"""
Module: plotter
---------------

A collection of classes for creating and working with plots.  This package
contains classes for handling scatter plots and plots of data on latitude and
longitude coordinates.

The plotting library is built on Matplotlib.pyplot and the mpltoolkit Basemap.
Instead of simply linking to these libraries, this package implements a small
subset of the capability of these libraries and integrates them into a user
interface.

the projections supported are orthographic, mercator, lambert conformal and
lambert azimuthal (equal area).
"""
